import datetime
from typing import Optional


def dttm_str_to_dttm(
    dttm_str: str, dttm_format: str = "%Y-%m-%dT%H:%M:%S.%f"
) -> Optional[datetime.datetime]:
    """Convert a datestamp string to a Date object"""

    if dttm_str and dttm_str != "":
        try:
            datetime_object = datetime.datetime.strptime(dttm_str, dttm_format)
            return datetime_object
        except Exception as e:
            return None
    else:
        return None


def dttm_to_dttm_str(
    dttm: datetime.datetime, dttm_format: str = "%Y-%m-%dT%H:%M:%S"
) -> Optional[str]:
    """Convert a datetime object to formatted string"""

    if dttm and isinstance(dttm, datetime.datetime):
        try:
            dttm_str = datetime.datetime.strftime(dttm, dttm_format)
            return dttm_str
        except Exception as e:
            return None
    else:
        return None


def current_datetime_utc() -> datetime.datetime:
    return datetime.datetime.now(datetime.timezone.utc)


def yesterday_datetime() -> datetime.datetime:
    return datetime.datetime.now() - datetime.timedelta(days=1)



# from typing import Optional
#
# from pak8.app.conf import Pak8AppConf
# from pak8.conf import Pak8Conf
# from phi.conf.phi_conf import PhiConf, PhiK8sData, PhiWsData
# from phi.utils.common import pprint_heading, pprint_status, pprint_subheading
# from phi.utils.log import logger
#
#
# def secho_pak8_service_status(pak8_svc: Pak8AppConf) -> None:
#
#     pprint_subheading("Service: {}".format(pak8_svc.name))
#     pprint_status("Type: {}".format(pak8_svc.type.value))
#     if pak8_svc.args:
#         pprint_status("Args: {}".format(pak8_svc.args))
#
#
# def secho_k8_status(ws_data: PhiWsData, config: PhiConf) -> None:
#
#     ws_pak8_conf: Optional[Pak8Conf] = ws_data.ws_pak8_conf
#
#     pprint_heading("K8s")
#     ws_k8s_data: Optional[PhiK8sData] = ws_data.ws_k8s_data
#     if ws_k8s_data:
#         logger.debug("kubeconfig available: {}".format(ws_k8s_data.kubeconfig_avl))
#         logger.debug("kubeconfig: {}".format(ws_k8s_data.kubeconfig_path))
#
#     # for pak8_svc in ws_pak8_conf.services:
#     #     secho_pak8_service_status(pak8_svc)
#     # resources_dir: Optional[Path] = config.get_k8s_resources_dir_path_by_ws_name(
#     #     ws_data.ws_name
#     # )
#     # if resources_dir:
#     #     pprint_status("Resources Saved to: {}".format(str(resources_dir)))



"""Phidata Kubectl Cli

This is the entrypoint for the `phi k` commands.
"""

from typing import Optional, cast

import typer

from phi.utils.cli_console import (
    print_error,
    print_info,
    print_available_workspaces,
    print_conf_not_available_msg,
    print_active_workspace_not_available,
)
from phi.workspace.ws_enums import WorkspaceEnv
from phi.utils.log import logger, set_log_level_to_debug

k8s_app = typer.Typer(
    name="k",
    help="Commands to manage your k8s resources",
    no_args_is_help=True,
    invoke_without_command=True,
)


@k8s_app.command(short_help="Save your K8s Resources")
def save(
    ws_filter: Optional[str] = typer.Argument(
        None,
        help="Filter which K8s configs to save. Format - ENV:NAME:TYPE",
        metavar="[filter]",
    ),
    env_filter: str = typer.Option(
        None,
        "-e",
        "--env",
        metavar="",
        help="The environment to use. Available Options: {}".format(
            WorkspaceEnv.values_list()
        ),
    ),
    name_filter: Optional[str] = typer.Option(
        None, "-n", "--name", metavar="", help="Filter using app name"
    ),
    type_filter: Optional[str] = typer.Option(
        None,
        "-t",
        "--type",
        metavar="",
        help="Filter using resource type",
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """
    Saves your k8s resources so you know exactly what is being deployed

    \b
    Examples:
    * `phi k save`      -> Save resources for the active workspace
    """

    from phi.conf.phi_conf import PhiConf, PhiWsData
    from phi.k8s.k8s_operator import save_k8s_resources

    if print_debug_log:
        set_log_level_to_debug()

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return

    active_ws_data: Optional[PhiWsData] = phi_conf.get_active_ws_data(refresh=True)
    if active_ws_data is None:
        print_active_workspace_not_available()
        avl_ws = phi_conf.available_ws
        if avl_ws:
            print_available_workspaces(avl_ws)
        return

    target_env_str: Optional[str] = None
    target_env: Optional[WorkspaceEnv] = None
    target_name: Optional[str] = None
    target_type: Optional[str] = None

    if ws_filter is not None:
        if not isinstance(ws_filter, str):
            raise TypeError(
                f"Invalid workspace filter. Expected: str, Received: {type(ws_filter)}"
            )
        filters = ws_filter.split(":")
        # logger.debug(f"filters: {filters}")
        num_filters = len(filters)
        if num_filters >= 1:
            target_env_str = filters[0]
        if num_filters >= 2:
            target_name = filters[2]
        if num_filters >= 3:
            target_type = filters[3]

    if target_env_str is None:
        target_env_str = env_filter
    if target_name is None:
        target_name = name_filter
    if target_type is None:
        target_type = type_filter

    if target_env_str is not None:
        if target_env_str.lower() not in WorkspaceEnv.values_list():
            print_error(
                f"{target_env_str} is not supported, please choose from: {WorkspaceEnv.values_list()}"
            )
            return
        target_env = cast(
            WorkspaceEnv,
            WorkspaceEnv.from_str(target_env_str),
        )

    logger.debug("Processing workspace")
    logger.debug(f"\ttarget_env: {target_env}")
    logger.debug(f"\ttarget_name: {target_name}")
    logger.debug(f"\ttarget_type: {target_type}")
    save_k8s_resources(
        ws_data=active_ws_data,
        target_env=target_env,
        target_name=target_name,
        target_type=target_type,
    )


# @app.command(short_help="Print your K8s Resources")
# def print(
#     refresh: bool = typer.Option(
#         False,
#         "-r",
#         "--refresh",
#         help="Refresh the workspace config, use this if you've just changed your phi-config.yaml",
#         show_default=True,
#     ),
#     type_filters: List[str] = typer.Option(
#         None, "-k", "--kind", help="Filter the K8s resources by kind"
#     ),
#     name_filters: List[str] = typer.Option(
#         None, "-n", "--name", help="Filter the K8s resources by name"
#     ),
# ):
#     """
#     Print your k8s resources so you know exactly what is being deploying
#
#     \b
#     Examples:
#     * `phi k print`      -> Print resources for the primary workspace
#     * `phi k print data` -> Print resources for the workspace named data
#     """
#
#     from phi import schemas
#     from phi.k8s import k8s_operator
#     from phi.conf.phi_conf import PhiConf
#
#     config: Optional[PhiConf] = PhiConf.get_saved_conf()
#     if not config:
#         conf_not_available_msg()
#         raise typer.Exit(1)
#
#     primary_ws: Optional[schemas.WorkspaceSchema] = config.primary_ws
#     if primary_ws is None:
#         primary_ws_not_available_msg()
#         raise typer.Exit(1)
#
#     k8s_operator.print_k8s_resources_as_yaml(
#         primary_ws, config, refresh, type_filters, name_filters
#     )
#
#
# @app.command(short_help="Apply your K8s Resources")
# def apply(
#     refresh: bool = typer.Option(
#         False,
#         "-r",
#         "--refresh",
#         help="Refresh the workspace config, use this if you've just changed your phi-config.yaml",
#         show_default=True,
#     ),
#     service_filters: List[str] = typer.Option(
#         None, "-s", "--svc", help="Filter the Services"
#     ),
#     type_filters: List[str] = typer.Option(
#         None, "-k", "--kind", help="Filter the K8s resources by kind"
#     ),
#     name_filters: List[str] = typer.Option(
#         None, "-n", "--name", help="Filter the K8s resources by name"
#     ),
# ):
#     """
#     Apply your k8s resources. You can filter the resources by services, kind or name
#
#     \b
#     Examples:
#     * `phi k apply`      -> Apply resources for the primary workspace
#     """
#
#     from phi import schemas
#     from phi.k8s import k8s_operator
#     from phi.conf.phi_conf import PhiConf
#
#     config: Optional[PhiConf] = PhiConf.get_saved_conf()
#     if not config:
#         conf_not_available_msg()
#         raise typer.Exit(1)
#
#     primary_ws: Optional[schemas.WorkspaceSchema] = config.primary_ws
#     if primary_ws is None:
#         primary_ws_not_available_msg()
#         raise typer.Exit(1)
#
#     k8s_operator.apply_k8s_resources(
#         primary_ws, config, refresh, service_filters, type_filters, name_filters
#     )
#
#
# @app.command(short_help="Get active K8s Objects")
# def get(
#     service_filters: List[str] = typer.Option(
#         None, "-s", "--svc", help="Filter the Services"
#     ),
#     type_filters: List[str] = typer.Option(
#         None, "-k", "--kind", help="Filter the K8s resources by kind"
#     ),
#     name_filters: List[str] = typer.Option(
#         None, "-n", "--name", help="Filter the K8s resources by name"
#     ),
# ):
#     """
#     Get active k8s resources.
#
#     \b
#     Examples:
#     * `phi k apply`      -> Get active resources for the primary workspace
#     """
#
#     from phi import schemas
#     from phi.k8s import k8s_operator
#     from phi.conf.phi_conf import PhiConf
#
#     config: Optional[PhiConf] = PhiConf.get_saved_conf()
#     if not config:
#         conf_not_available_msg()
#         raise typer.Exit(1)
#
#     primary_ws: Optional[schemas.WorkspaceSchema] = config.primary_ws
#     if primary_ws is None:
#         primary_ws_not_available_msg()
#         raise typer.Exit(1)
#
#     k8s_operator.print_active_k8s_resources(
#         primary_ws, config, service_filters, type_filters, name_filters
#     )
#
#
# if __name__ == "__main__":
#     app()



"""Phi Devbox Cli

This is the entrypoint for the `phi dx` commands
"""


from typing import Optional

import typer

from phi.utils.cli_console import (
    print_error,
    print_info,
    print_conf_not_available_msg,
    print_active_workspace_not_available,
    print_available_workspaces,
)
from phi.utils.log import logger, set_log_level_to_debug
from phi.workspace.ws_enums import WorkspaceEnv

dx_app = typer.Typer(
    name="dx",
    short_help="Run commands on your devbox or databox",
    no_args_is_help=True,
    add_completion=False,
    invoke_without_command=True,
    options_metavar="\b",
    subcommand_metavar="<command>",
)


@dx_app.command(
    short_help="Run command in devbox or databox",
    options_metavar="\b",
    no_args_is_help=True,
)
def run(
    command: str = typer.Argument(..., help="Command to run.", metavar="[cmd]"),
    env_filter: Optional[str] = typer.Option(
        None,
        "-e",
        "--env",
        metavar="",
        help="The environment to run the command in. Default: docker. Available Options: {}".format(
            WorkspaceEnv.values_list()
        ),
    ),
    app_name: Optional[str] = typer.Option(
        None,
        "-a",
        metavar="",
        help="The App to run the workflow in. Default: `devbox` for docker and `databox` for K8s.",
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    from phi.conf.phi_conf import PhiConf, PhiWsData
    from phi.devbox.devbox_operator import run_command

    if print_debug_log:
        set_log_level_to_debug()

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return

    active_ws_data: Optional[PhiWsData] = phi_conf.get_active_ws_data(refresh=True)
    if active_ws_data is None:
        print_active_workspace_not_available()
        avl_ws = phi_conf.available_ws
        if avl_ws:
            print_available_workspaces(avl_ws)
        return

    target_env: Optional[WorkspaceEnv] = None
    target_app: Optional[str] = app_name

    target_env_str = env_filter or active_ws_data.ws_config.default_env or "dev"
    try:
        target_env = WorkspaceEnv.from_str(target_env_str)
    except Exception as e:
        print_error(e)
        print_error(
            f"{target_env_str} is not supported, please choose from: {WorkspaceEnv.values_list()}"
        )
        return

    logger.debug("Running workflow")
    logger.debug(f"\tcommand      : {command}")
    logger.debug(f"\ttarget_env   : {target_env}")
    logger.debug(f"\ttarget_app   : {target_app}")
    run_command(
        command=command,
        ws_data=active_ws_data,
        target_env=target_env,
        target_app=target_app,
    )



"""Phi Workspace Cli

This is the entrypoint for the `phi ws` commands
"""

from typing import Optional, cast

import typer

from phi.utils.cli_console import (
    print_error,
    print_info,
    print_available_workspaces,
    print_conf_not_available_msg,
    print_active_workspace_not_available,
)
from phi.utils.log import logger, set_log_level_to_debug
from phi.workspace.ws_enums import (
    WorkspaceEnv,
    WorkspaceStarterTemplate,
    WorkspaceInfra,
)

ws_app = typer.Typer(
    name="ws",
    short_help="Commands to manage your workspaces",
    help="""\b
Use `phi ws <command>` to create, setup or delete your phidata workspace.
Run `phi ws <command> --help` for more info.
""",
    no_args_is_help=True,
    add_completion=False,
    invoke_without_command=True,
    options_metavar="\b",
    subcommand_metavar="<command>",
)


@ws_app.command(short_help="Create a new phidata workspace in the current directory.")
def init(
    ws_name: str = typer.Option(None, "-ws", help="Name of the new workspace"),
    template: str = typer.Option(
        "docker",
        "-t",
        "--template",
        help="Choose a starter template which comes pre-populated with defaults. Available Options: {}".format(
            ", ".join(WorkspaceStarterTemplate.values_list())
        ),
        show_default=True,
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """\b
    Creates a new workspace in the current directory using the selected starter template
    [default: docker]

    \b
    Examples:
    $ phi ws init                -> Create a new workspace
    $ phi ws init -ws data        -> Create a workspace named data using the default starter template (docker)
    $ phi ws init -ws data -t aws -> Create a workspace named data using the `aws` starter template
    """
    from phi.workspace.ws_operator import (
        create_new_workspace,
        initialize_workspace,
    )

    if print_debug_log:
        set_log_level_to_debug()

    if ws_name is None:
        initialize_workspace()
        return

    if (
        template is None
        or template.lower() not in WorkspaceStarterTemplate.values_list()
    ):
        print_error(
            f"{template} is not a supported template, please choose from: {WorkspaceStarterTemplate.values_list()}"
        )
        return
    _template: WorkspaceStarterTemplate = cast(
        WorkspaceStarterTemplate,
        WorkspaceStarterTemplate.from_str(template),
    )

    create_new_workspace(ws_name, _template)


@ws_app.command(short_help="Setup a phidata workspace")
def setup(
    ws_name: str = typer.Option(None, "-ws", help="Name of the workspace to setup"),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """\b
    Setup the current directory as a phidata workspace.
    This command can be run from within the workspace directory
        OR with a -ws flag to setup another workspace.

    \b
    Examples:
    $ `phi ws setup`         -> Setup the current directory as a phidata workspace
    $ `phi ws setup -ws data` -> Setup the workspace named idata
    """
    from pathlib import Path
    from phi.workspace.ws_operator import setup_workspace

    if print_debug_log:
        set_log_level_to_debug()

    # By default, we assume this command is run from the workspace directory
    if ws_name is None:
        # If the user does not provide a ws_name, that implies `phi ws setup` is ran from
        # the workspace directory.
        ws_path: Path = Path(".").resolve()
        setup_workspace(ws_path)
    else:
        # If the user provides a workspace name manually, we find the dir for that ws and set it up
        from phi.conf.phi_conf import PhiConf

        phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
        if not phi_conf:
            print_conf_not_available_msg()
            return

        _ws_path: Optional[Path] = phi_conf.get_ws_dir_path_by_name(ws_name)
        if _ws_path is None:
            print_error(f"Could not find workspace {ws_name}")
            avl_ws = phi_conf.available_ws
            if avl_ws:
                print_available_workspaces(avl_ws)
            return
        setup_workspace(_ws_path)


# @ws_app.command(short_help="Show steps to setup a git repo for the ws")
# def git(
#     ws_name: str = typer.Option(None, "-ws", help="[Optional] Name for the workspace"),
#     print_debug_log: bool = typer.Option(
#         False,
#         "-d",
#         "--debug",
#         help="Print debug logs.",
#     ),
# ):
#     """
#     Print steps to setup a remote git repo for the active workspace.
#
#     \b
#     Examples:
#     $ `phi ws git`
#     $ `phi ws git -ws data`
#     """
#     from pathlib import Path
#     from phi.enums.user_enums import VersionControlProviderEnum
#     from phi.workspace.ws_operator import print_git_setup
#
#     from phi.conf.phi_conf import PhiConf
#
#     phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
#     if not phi_conf:
#         print_conf_not_available_msg()
#         return
#
#     _ws_name: Optional[str] = None
#     ws_dir_path: Optional[Path] = None
#     if ws_name:
#         _ws_name = ws_name
#         ws_dir_path = phi_conf.get_ws_dir_path_by_name(_ws_name)
#     else:
#         # print steps for the active workspace
#         _ws_name = phi_conf.active_ws_name
#         if _ws_name is None:
#             print_info(
#                 "Primary workspace not available, searching current directory for a workspace"
#             )
#             ws_dir_path = Path(".").resolve()
#             _ws_name = phi_conf.get_ws_name_by_path(ws_dir_path)
#         else:
#             ws_data = phi_conf.get_ws_data_by_name(_ws_name)
#             if ws_data is not None and ws_data.ws_dir_path is not None:
#                 ws_dir_path = ws_data.ws_dir_path
#
#     if _ws_name is None or ws_dir_path is None:
#         print_error(f"Could not find workspace at {ws_dir_path}")
#         avl_ws = phi_conf.available_ws
#         if avl_ws:
#             print_available_workspaces(avl_ws)
#         return
#
#     version_control_provider = (
#         phi_conf.user.version_control_provider
#         if phi_conf.user and phi_conf.user.version_control_provider
#         else VersionControlProviderEnum.GITHUB
#     )
#     print_git_setup(
#         ws_name=_ws_name,
#         ws_dir_path=ws_dir_path,
#         version_control_provider=version_control_provider,
#     )


@ws_app.command(
    short_help="Deploy your active workspace",
    options_metavar="\b",
)
def up(
    ws_filter: Optional[str] = typer.Argument(
        None,
        help="Filter the resources to deploy. Format - ENV:NAME:TYPE:INFRA",
        metavar="[filter]",
    ),
    env_filter: Optional[str] = typer.Option(
        None,
        "-e",
        "--env",
        metavar="",
        help="Filter the environment to deploy. Available Options: {}".format(
            WorkspaceEnv.values_list()
        ),
    ),
    infra_filter: Optional[str] = typer.Option(
        None,
        "-i",
        "--infra",
        metavar="",
        help="Filter the infrastructure to deploy. Available Options: {}".format(
            WorkspaceInfra.values_list()
        ),
    ),
    name_filter: Optional[str] = typer.Option(
        None, "-n", "--name", metavar="", help="Filter using resource name"
    ),
    type_filter: Optional[str] = typer.Option(
        None,
        "-t",
        "--type",
        metavar="",
        help="Filter using resource type",
    ),
    dry_run: bool = typer.Option(
        False,
        "-dr",
        "--dry-run",
        help="Print which resources will be deployed and exit.",
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """
    \b
    Deploys the active workspace.
    Filters can be used to limit deployment by
        - env (dev, stg, prd)
        - infra (docker, aws, k8s)
        - Resource Name
        - Resource Type
    \b
    Filters can be provided as a single argument or using multiple options.
    Examples:
    \b
    $ `phi ws up`           -> Deploy default resources
    $ `phi ws up dev`       -> Deploy all dev resources
    $ `phi ws up prd`       -> Deploy all prd resources
    $ `phi ws up prd:::aws` -> Deploy all prd aws resources
    $ `phi ws up prd:s3`    -> Deploy prd resources with name s3
    """
    from phi.conf.phi_conf import PhiConf, PhiWsData
    from phi.workspace.ws_operator import deploy_workspace

    if print_debug_log:
        set_log_level_to_debug()

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return

    active_ws_data: Optional[PhiWsData] = phi_conf.get_active_ws_data(refresh=True)
    if active_ws_data is None:
        print_active_workspace_not_available()
        avl_ws = phi_conf.available_ws
        if avl_ws:
            print_available_workspaces(avl_ws)
        return

    target_env_str: Optional[str] = None
    target_infra_str: Optional[str] = None
    target_env: Optional[WorkspaceEnv] = None
    target_infra: Optional[WorkspaceInfra] = None
    target_name: Optional[str] = None
    target_type: Optional[str] = None

    # derive env/infra/name/type from ws_filter
    if ws_filter is not None:
        if not isinstance(ws_filter, str):
            raise TypeError(
                f"Invalid workspace filter. Expected: str, Received: {type(ws_filter)}"
            )
        filters = ws_filter.split(":")
        # logger.debug(f"filters: {filters}")
        num_filters = len(filters)
        if num_filters >= 1:
            target_env_str = filters[0]
        if num_filters >= 2:
            target_name = filters[1]
        if num_filters >= 3:
            target_type = filters[2]
        if num_filters >= 4:
            target_infra_str = filters[3]

    # derive env/infra/name/type from command options
    if target_env_str is None:
        target_env_str = env_filter
    if target_infra_str is None:
        target_infra_str = infra_filter
    if target_name is None:
        target_name = name_filter
    if target_type is None:
        target_type = type_filter

    # derive env/infra/name/type from defaults
    if target_env_str is None:
        target_env_str = active_ws_data.ws_config.default_env
    if target_infra_str is None:
        target_infra_str = active_ws_data.ws_config.default_infra

    if target_env_str is not None:
        if target_env_str.lower() not in WorkspaceEnv.values_list():
            print_error(
                f"{target_env_str} is not supported, please choose from: {WorkspaceEnv.values_list()}"
            )
            return
        target_env = cast(
            WorkspaceEnv,
            WorkspaceEnv.from_str(target_env_str),
        )

    if target_infra_str is not None:
        if target_infra_str.lower() not in WorkspaceInfra.values_list():
            print_error(
                f"{target_infra_str} is not supported, please choose from: {WorkspaceInfra.values_list()}"
            )
            return
        target_infra = cast(
            WorkspaceInfra,
            WorkspaceInfra.from_str(target_infra_str),
        )

    logger.debug("Deploying workspace")
    logger.debug(f"\ttarget_env  : {target_env}")
    logger.debug(f"\ttarget_infra: {target_infra}")
    logger.debug(f"\ttarget_name : {target_name}")
    logger.debug(f"\ttarget_type : {target_type}")
    deploy_workspace(
        ws_data=active_ws_data,
        target_env=target_env,
        target_infra=target_infra,
        target_name=target_name,
        target_type=target_type,
        dry_run=dry_run,
    )


@ws_app.command(short_help="Shutdown your active workspace - does not delete any data")
def down(
    ws_filter: Optional[str] = typer.Argument(
        None,
        help="Filter the resources to shut down. Format - ENV:NAME:TYPE:INFRA",
        metavar="[filter]",
    ),
    env_filter: str = typer.Option(
        None,
        "-e",
        "--env",
        metavar="",
        help="Filter the environment to shut down. Available Options: {}".format(
            WorkspaceEnv.values_list()
        ),
    ),
    infra_filter: str = typer.Option(
        None,
        "-i",
        "--infra",
        metavar="",
        help="Filter the infrastructure to shut down. Available Options: {}".format(
            WorkspaceInfra.values_list()
        ),
    ),
    name_filter: Optional[str] = typer.Option(
        None, "-n", "--name", metavar="", help="Filter using resource name"
    ),
    type_filter: Optional[str] = typer.Option(
        None,
        "-t",
        "--type",
        metavar="",
        help="Filter using resource type",
    ),
    dry_run: bool = typer.Option(
        False,
        "-dr",
        "--dry-run",
        help="Print which resources will be shut down and exit.",
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """
    \b
    Shuts down the active workspace but does not delete it.

    \b
    Examples:
    $ `phi ws down`
    """
    from phi.conf.phi_conf import PhiConf, PhiWsData
    from phi.workspace.ws_operator import shutdown_workspace

    if print_debug_log:
        set_log_level_to_debug()

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return

    active_ws_data: Optional[PhiWsData] = phi_conf.get_active_ws_data(refresh=True)
    if active_ws_data is None:
        print_active_workspace_not_available()
        avl_ws = phi_conf.available_ws
        if avl_ws:
            print_available_workspaces(avl_ws)
        return

    target_env_str: Optional[str] = None
    target_infra_str: Optional[str] = None
    target_env: Optional[WorkspaceEnv] = None
    target_infra: Optional[WorkspaceInfra] = None
    target_name: Optional[str] = None
    target_type: Optional[str] = None

    # derive env/infra/name/type from ws_filter
    if ws_filter is not None:
        if not isinstance(ws_filter, str):
            raise TypeError(
                f"Invalid workspace filter. Expected: str, Received: {type(ws_filter)}"
            )
        filters = ws_filter.split(":")
        # logger.debug(f"filters: {filters}")
        num_filters = len(filters)
        if num_filters >= 1:
            target_env_str = filters[0]
        if num_filters >= 2:
            target_name = filters[1]
        if num_filters >= 3:
            target_type = filters[2]
        if num_filters >= 4:
            target_infra_str = filters[3]

    # derive env/infra/name/type from command options
    if target_env_str is None:
        target_env_str = env_filter
    if target_infra_str is None:
        target_infra_str = infra_filter
    if target_name is None:
        target_name = name_filter
    if target_type is None:
        target_type = type_filter

    # derive env/infra/name/type from defaults
    if target_env_str is None:
        target_env_str = active_ws_data.ws_config.default_env
    if target_infra_str is None:
        target_infra_str = active_ws_data.ws_config.default_infra

    if target_env_str is not None:
        if target_env_str.lower() not in WorkspaceEnv.values_list():
            print_error(
                f"{target_env_str} is not supported, please choose from: {WorkspaceEnv.values_list()}"
            )
            return
        target_env = cast(
            WorkspaceEnv,
            WorkspaceEnv.from_str(target_env_str),
        )

    if target_infra_str is not None:
        if target_infra_str.lower() not in WorkspaceInfra.values_list():
            print_error(
                f"{target_infra_str} is not supported, please choose from: {WorkspaceInfra.values_list()}"
            )
            return
        target_infra = cast(
            WorkspaceInfra,
            WorkspaceInfra.from_str(target_infra_str),
        )

    logger.debug("Shutting down workspace")
    logger.debug(f"\ttarget_env: {target_env}")
    logger.debug(f"\ttarget_infra: {target_infra}")
    logger.debug(f"\ttarget_name: {target_name}")
    logger.debug(f"\ttarget_type: {target_type}")
    shutdown_workspace(
        ws_data=active_ws_data,
        target_env=target_env,
        target_infra=target_infra,
        target_name=target_name,
        target_type=target_type,
        dry_run=dry_run,
    )


@ws_app.command(short_help="Restart your active workspace")
def restart(
    ws_filter: Optional[str] = typer.Argument(
        None,
        help="Filter the resources to restart. Format - ENV:NAME:TYPE:INFRA",
        metavar="[filter]",
    ),
    env_filter: str = typer.Option(
        None,
        "-e",
        "--env",
        metavar="",
        help="Filter the environment to restart. Available Options: {}".format(
            WorkspaceEnv.values_list()
        ),
    ),
    infra_filter: str = typer.Option(
        None,
        "-i",
        "--infra",
        metavar="",
        help="Filter the infrastructure to restart. Available Options: {}".format(
            WorkspaceInfra.values_list()
        ),
    ),
    name_filter: Optional[str] = typer.Option(
        None, "-n", "--name", metavar="", help="Filter using resource name"
    ),
    type_filter: Optional[str] = typer.Option(
        None,
        "-t",
        "--type",
        metavar="",
        help="Filter using resource type",
    ),
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """
    \b
    Restarts the active workspace. i.e. runs `phi ws down` and then `phi ws up`.

    \b
    Examples:
    $ `phi ws restart`
    """
    from time import sleep

    down(
        ws_filter=ws_filter,
        env_filter=env_filter,
        infra_filter=infra_filter,
        name_filter=name_filter,
        type_filter=type_filter,
        dry_run=False,
        print_debug_log=print_debug_log,
    )
    print_info("Sleeping for 2 seconds..")
    sleep(2)
    up(
        ws_filter=ws_filter,
        env_filter=env_filter,
        infra_filter=infra_filter,
        name_filter=name_filter,
        type_filter=type_filter,
        dry_run=False,
        print_debug_log=print_debug_log,
    )
    print_info("Workspace restarted!")


@ws_app.command(short_help="Prints active workspace config")
def config(
    print_debug_log: bool = typer.Option(
        False,
        "-d",
        "--debug",
        help="Print debug logs.",
    ),
):
    """\b
    Prints the active workspace config

    \b
    Examples:
    $ `phi ws config`         -> Print the active workspace config
    """
    from phi.conf.phi_conf import PhiConf, PhiWsData

    if print_debug_log:
        set_log_level_to_debug()

    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return

    active_ws_data: Optional[PhiWsData] = phi_conf.get_active_ws_data(refresh=True)
    if active_ws_data is None:
        print_active_workspace_not_available()
        avl_ws = phi_conf.available_ws
        if avl_ws:
            print_available_workspaces(avl_ws)
        return

    active_ws_data.print_to_cli()



# import datetime
# from typing import Any, Dict, Optional
#
# from pydantic import BaseModel
#
# from pak8.gcp.conf import Pak8GCPConf
# from phi.gcp.gcp_schemas import GCPProjectSchema, GKEClusterSchema
# from phi.utils.dttm import current_datetime_utc
#
#
# class PhiGcpData(BaseModel):
#     """The PhiGcpData schema contains all information required for a phidata workspace
#     to interface with the Google Cloud Platform.
#
#     Field Descriptions:
#     gcp_project: GCPProjectSchema schema.
#     gke_cluster: GKEClusterSchema schema.
#     gcloud_default_creds_avl: If the gcloud sdk is available & authenticated on this machine.
#         Then we can use the application-default-credentials: https://googleapis.dev/python/google-auth/latest/user-guide.html#application-default-credentials
#     gcp_svc_account: If this user is the workspace owner, they would probably create the service account
#         for phidata. If they do, then we can store details about that service account in gcp_svc_account
#     """
#
#     gcp_project_id: str
#     gcp_svc_account: Optional[Dict[str, Any]] = None
#     pak8_gcp_conf: Optional[Pak8GCPConf] = None
#     gcloud_default_creds_avl: bool = False
#     last_update_ts: datetime.datetime = current_datetime_utc()
#     gcp_project_schema: Optional[GCPProjectSchema] = None
#     gke_cluster_schema: Optional[GKEClusterSchema] = None



from typing import Dict

from phi.workspace.ws_enums import WorkspaceStarterTemplate

template_to_repo_map: Dict[WorkspaceStarterTemplate, str] = {
    WorkspaceStarterTemplate.docker: "https://github.com/phidata-public/phidata_starter_aws.git",
    WorkspaceStarterTemplate.aws: "https://github.com/phidata-public/phidata_starter_aws.git",
}



from pathlib import Path
from typing import List, Optional, cast, Dict, Tuple, Any

from phidata.infra.base import InfraConfig
from phidata.infra.k8s.api_client import K8sApiClient
from phidata.infra.k8s.config import K8sConfig
from phidata.infra.k8s.exceptions import K8sConfigException
from phidata.infra.k8s.manager import K8sManager
from phidata.infra.k8s.resource.types import K8sResourceType
from phidata.workspace import WorkspaceConfig
from phidata.workflow import Workflow
from phidata.types.context import PathContext, RunContext

from phi.workspace.phi_ws_data import PhiWsData
from phi.workspace.ws_enums import WorkspaceEnv
from phi.types.run_status import RunStatus
from phi.utils.filesystem import delete_files_in_dir
from phi.utils.cli_console import (
    print_error,
    print_heading,
    print_info,
)
from phi.utils.log import logger


def deploy_k8s_config(
    config: K8sConfig,
    name_filter: Optional[str] = None,
    type_filter: Optional[str] = None,
    dry_run: Optional[bool] = False,
) -> bool:

    # Step 1: Get the K8sManager
    k8s_manager: K8sManager = config.get_k8s_manager()
    if k8s_manager is None:
        raise K8sConfigException("K8sManager unavailable")

    # Step 3: If dry_run, print the resources and return True
    if dry_run:
        k8s_manager.create_resources_dry_run(
            name_filter=name_filter, type_filter=type_filter
        )
        return True

    # Step 3: Create resources
    env = config.env
    print_heading(
        "Deploying K8s config{}\n".format(f" for env: {env}" if env is not None else "")
    )
    try:
        success: bool = k8s_manager.create_resources(
            name_filter=name_filter, type_filter=type_filter
        )
        if not success:
            return False
    except Exception:
        raise

    # Step 4: Validate resources are created
    resource_creation_valid: bool = k8s_manager.validate_resources_are_created(
        name_filter=name_filter, type_filter=type_filter
    )
    if not resource_creation_valid:
        logger.error("K8sResource creation could not be validated")
        return False

    print_info("K8s config deployed")
    return True


def shutdown_k8s_config(
    config: K8sConfig,
    name_filter: Optional[str] = None,
    type_filter: Optional[str] = None,
    dry_run: Optional[bool] = False,
) -> bool:

    # Step 1: Get the K8sManager
    k8s_manager: K8sManager = config.get_k8s_manager()
    if k8s_manager is None:
        raise K8sConfigException("K8sManager unavailable")

    # Step 2: If dry_run, print the resources and return True
    if dry_run:
        k8s_manager.delete_resources_dry_run(
            name_filter=name_filter, type_filter=type_filter
        )
        return True

    # Step 3: Delete resources
    env = config.env
    print_heading(
        "Shutting down K8s config{}\n".format(
            f" for env: {env}" if env is not None else ""
        )
    )
    try:
        success: bool = k8s_manager.delete_resources(
            name_filter=name_filter, type_filter=type_filter
        )
        if not success:
            return False
    except Exception:
        raise

    # Step 4: Validate resources are deleted
    resources_deletion_valid: bool = k8s_manager.validate_resources_are_deleted(
        name_filter=name_filter, type_filter=type_filter
    )
    if not resources_deletion_valid:
        logger.error("K8sResource deletion could not be validated")
        return False

    print_info("K8s config shutdown")
    return True


def save_k8s_resources(
    ws_data: PhiWsData,
    target_env: Optional[WorkspaceEnv] = None,
    target_name: Optional[str] = None,
    target_type: Optional[str] = None,
) -> None:
    """Saves the K8s resources"""

    if ws_data is None or ws_data.ws_config is None:
        print_error("WorkspaceConfig invalid")
        return

    ws_name: str = ws_data.ws_name
    # print_heading(f"Building K8s manifests for {ws_name}")

    # Step 1: Get the K8sConfigs to build
    configs_to_build: List[InfraConfig] = []
    if target_env is None:
        if ws_data.ws_config.k8s is not None:
            configs_to_build.extend(ws_data.ws_config.k8s)
    else:
        for config in ws_data.ws_config.k8s:
            if config.env is None:
                continue
            if config.env.lower() not in WorkspaceEnv.values_list():
                print_error(f"{config.env} not supported")
                continue
            config_env = WorkspaceEnv.from_str(config.env)
            if config_env == target_env:
                configs_to_build.append(config)

    num_configs_to_build = len(configs_to_build)
    num_configs_built = 0
    for config in configs_to_build:
        env = config.env
        print_heading(
            "\nBuilding K8s manifests{}\n".format(
                f" for env: {env}" if env is not None else ""
            )
        )
        if not isinstance(config, InfraConfig):
            print_error(f"{config} is not an instance of InfraConfig")
            continue
        if not isinstance(config, K8sConfig):
            print_error(f"{config} is not an instance of K8sConfig")
            continue
        if not config.is_valid():
            print_error("K8sConfig invalid")
            continue
        if not config.enabled:
            logger.debug(
                f"{config.__class__.__name__} for env {config.__class__.env} disabled"
            )
            num_configs_built += 1
            continue

        # NOTE: Very important step
        # If the paths arent updated then the Manager object cannot be created
        # Update the workspace_root_path and workspace_config_file_path
        config.workspace_dir_path = ws_data.ws_dir_path
        config.workspace_config_file_path = ws_data.ws_config_file_path

        # Step 2: Get the K8sManager
        k8s_manager: K8sManager = config.get_k8s_manager()
        if k8s_manager is None:
            raise K8sConfigException("K8sManager unavailable")

        # Step 3: Get the K8sResources
        k8s_resources: Dict[str, List[K8sResourceType]] = k8s_manager.read_resources(
            name_filter=target_name, type_filter=target_type
        )

        # Step 4: Prep the directory to write to
        workspace_config_file_path: Optional[Path] = ws_data.ws_config_file_path
        if workspace_config_file_path is None:
            print_error("workspace_config_file_path invalid")
            continue
        workspace_config_dir: Path = workspace_config_file_path.parent.resolve()
        if workspace_config_dir is None:
            print_error("workspace_config_dir invalid")
            continue

        resources_dir: Path = workspace_config_dir.joinpath(config.resources_dir)
        if config.env is not None:
            resources_dir = resources_dir.joinpath(config.env)
        if config.args.name is not None:
            resources_dir = resources_dir.joinpath(config.args.name)

        # delete directory to save resources if needed
        if resources_dir.exists():
            print_info(f"Deleting {str(resources_dir)}")
            if resources_dir.is_file():
                resources_dir.unlink()
            elif resources_dir.is_dir():
                delete_files_in_dir(resources_dir)
        print_info(f"Saving to {str(resources_dir)}")
        resources_dir.mkdir(exist_ok=True, parents=True)

        for rg_name, resources_list in k8s_resources.items():
            print_info(f"Processing {rg_name}")
            rg_dir: Path = resources_dir.joinpath(rg_name)
            rg_dir.mkdir(exist_ok=True)
            for resource in resources_list:
                if resource is not None:
                    resource_name = resource.get_resource_name()
                    resource_file: Path = rg_dir.joinpath(f"{resource_name}.yaml")
                    try:
                        manifest_yaml = resource.get_k8s_manifest_yaml(
                            default_flow_style=False
                        )  # use default_style='"' for debugging
                        if manifest_yaml is not None:
                            logger.debug(f"Writing {str(resource_file)}")
                            resource_file.write_text(manifest_yaml)
                    except Exception as e:
                        logger.error(f"Could not parse {resource_name}")
                        continue
        num_configs_built += 1

    print_info(f"\n# Configs built: {num_configs_built}/{num_configs_to_build}")



# Need to be sure that we dont import anything which may lead to circular imports
import pickle
from pathlib import Path
from typing import Any, List, Optional, Type


def isinstanceany(obj: Any, class_list: List[Type]) -> bool:
    for cls in class_list:
        if isinstance(obj, cls):
            return True
    return False


def unpickle_object_from_file(
    file_path: Path, verify_class: Optional[Any] = None
) -> Any:
    """Reads the contents of file_path and unpickles the binary content into an object.
    If verify_class is provided, checks if the object is an instance of that class.
    """
    from phi.utils.log import logger

    _obj = None
    # logger.debug(f"Reading {file_path}")
    if file_path.exists() and file_path.is_file():
        _obj = pickle.load(file_path.open("rb"))

    # if verify_class is not None:
    #     logger.debug(f"Verifying if obj matches {verify_class}")

    if _obj and verify_class and not isinstance(_obj, verify_class):
        logger.error(f"Unpickled object does not match {verify_class}")
        _obj = None

    return _obj


def pickle_object_to_file(obj: Any, file_path: Path) -> Any:
    """Pickles and saves object to file_path"""

    _obj_parent = file_path.parent
    if not _obj_parent.exists():
        _obj_parent.mkdir(parents=True, exist_ok=True)
    pickle.dump(obj, file_path.open("wb"))


def str_to_int(inp: str) -> Optional[int]:
    """
    Safely converts a string value to integer.
    Args:
        inp: input string

    Returns: input string as int if possible, None if not
    """
    try:
        val = int(inp)
        return val
    except Exception:
        return None


def is_empty(val: Any) -> bool:
    if val is None or len(val) == 0 or val == "":
        return True
    return False



from dataclasses import dataclass
from typing import Optional


@dataclass
class BackendApi:

    # /auth paths
    cli_auth: str = "auth/cliauth"
    sign_in: str = "auth/signin"
    sign_up: str = "auth/signup"
    authenticate: str = "auth/authenticate"
    ping: str = "auth/ping"

    # /user paths
    user_read: str = "user/read"
    user_update: str = "user/update"

    # /workspace paths
    workspace_update: str = "workspace/update"
    workspace_read_primary: str = "workspace/read/primary"
    workspaces_read_available: str = "workspace/read/available"
    workspace_upsert_from_cli: str = "workspace/upsertfromcli"


def get_authenticated_client():
    """Returns an instance of httpx.Client which with preconfigured auth and base url"""

    from httpx import Client

    from phi.conf.constants import (
        APP_NAME,
        APP_VERSION,
        BACKEND_API_URL,
        PHI_SESSION_COOKIE_KEY,
    )
    from phi.conf.phi_cli_creds import read_session_cookie

    try:
        session_cookie: Optional[str] = read_session_cookie()
    except Exception:
        # print_error("Could not authenticating user. Please run `phi auth` to fix")
        return None

    if session_cookie is None:
        return None
    return Client(
        base_url=BACKEND_API_URL,
        headers={"user-agent": f"{APP_NAME}/{APP_VERSION}"},
        cookies={PHI_SESSION_COOKIE_KEY: session_cookie},
        timeout=60,
    )



from typing import Any, Dict, List, Optional, Union

import httpx

from phi.backend_api.api_client import BackendApi, get_authenticated_client
from phi.backend_api.api_exceptions import CliAuthException
from phi.backend_api.api_utils import headers, response_is_invalid
from phi.enums.response_enums import BackendResponseStatus
from phi.schemas.auth_schemas import EmailPasswordSignInSchema
from phi.schemas.user_schemas import UserSchema
from phi.utils.cli_console import log_network_error_msg, print_error
from phi.utils.log import logger


def is_user_authenticated() -> bool:
    logger.debug("++**++ Authenticating user")
    authenticated_client = get_authenticated_client()
    if authenticated_client is None:
        logger.debug("Session not available")
        return False

    logger.debug("Cookie Jar: {}".format(authenticated_client.cookies.jar))
    with authenticated_client as api:
        try:
            r: httpx.Response = api.get(BackendApi.authenticate)
            if response_is_invalid(r):
                return False
        except httpx.NetworkError as e:
            log_network_error_msg()
            return False

        logger.debug("url: {}".format(r.url))
        logger.debug("status: {}".format(r.status_code))
        logger.debug("headers: {}".format(r.headers))
        logger.debug("cookies: {}".format(r.cookies))

        _data: Union[Dict[Any, Any], List[Any]] = r.json()
        logger.debug("_data: {}".format(_data))

        if _data is None or not isinstance(_data, dict):
            print_error("Could not parse response")
            return False

        if _data.get("status", "FAIL") == BackendResponseStatus.SUCCESS.value:
            return True
        return False


def authenticate_and_get_user_schema(
    tmp_access_token: str,
) -> Optional[UserSchema]:
    from phi.conf.constants import BACKEND_API_URL, PHI_SESSION_COOKIE_KEY
    from phi.conf.phi_cli_creds import save_session_cookie

    with httpx.Client(base_url=BACKEND_API_URL, headers=headers, timeout=None) as api:
        logger.debug(f"authenticate_and_get_user_schema: {tmp_access_token}")
        try:
            r: httpx.Response = api.post(
                BackendApi.cli_auth, json={"access_token": tmp_access_token}
            )
            if response_is_invalid(r):
                return None
        except httpx.NetworkError as e:
            log_network_error_msg()
            return None

        logger.debug("status: {}".format(r.status_code))
        logger.debug("headers: {}".format(r.headers))
        logger.debug("cookies: {}".format(r.cookies))
        logger.debug("url: {}".format(r.url))
        logger.debug("json: {}".format(r.json()))

        phidata_session_cookie = r.cookies.get(PHI_SESSION_COOKIE_KEY)
        if phidata_session_cookie is None:
            print_error("Could not authenticate")
            return None

        user_data = r.json()
        if not isinstance(user_data, dict):
            raise CliAuthException("Could not parse response after authentication")

        current_user: UserSchema = UserSchema.from_dict(user_data)
        if current_user is not None:
            save_session_cookie(phidata_session_cookie)
            return current_user
        return None


def sign_in_user(
    sign_in_data: EmailPasswordSignInSchema,
) -> Optional[UserSchema]:
    logger.debug("--o-o-- Sign in user")
    from phi.conf.constants import BACKEND_API_URL, PHI_SESSION_COOKIE_KEY
    from phi.conf.phi_cli_creds import save_session_cookie

    with httpx.Client(base_url=BACKEND_API_URL, headers=headers, timeout=None) as api:
        try:
            r: httpx.Response = api.post(BackendApi.sign_in, json=sign_in_data.dict())
            if response_is_invalid(r):
                return None
        except httpx.NetworkError as e:
            log_network_error_msg()
            return None

        # logger.debug("status: {}".format(r.status_code))
        # logger.debug("headers: {}".format(r.headers))
        # logger.debug("cookies: {}".format(r.cookies))
        # logger.debug("url: {}".format(r.url))
        # logger.debug("json: {}".format(r.json()))

        phidata_session_cookie = r.cookies.get(PHI_SESSION_COOKIE_KEY)
        if phidata_session_cookie is None:
            print_error("Could not authenticate")
            return None

        user_data = r.json()
        if not isinstance(user_data, dict):
            raise CliAuthException("Could not parse response after authentication")

        current_user: UserSchema = UserSchema.from_dict(user_data)
        if current_user is not None:
            save_session_cookie(phidata_session_cookie)
            return current_user
        return None


def auth_ping() -> bool:
    """Returns true if phidata servers are reachable"""

    logger.debug("--o-o-- Pinging backend")
    from phi.conf.constants import BACKEND_API_URL

    with httpx.Client(base_url=BACKEND_API_URL, headers=headers, timeout=None) as api:
        try:
            r: httpx.Response = api.get(BackendApi.ping)
            if response_is_invalid(r):
                return False
        except httpx.NetworkError as e:
            logger.exception(e)
            log_network_error_msg()
            return False

        # logger.debug("status: {}".format(r.status_code))
        # logger.debug("headers: {}".format(r.headers))
        # logger.debug("cookies: {}".format(r.cookies))
        # logger.debug("url: {}".format(r.url))
        # logger.debug("json: {}".format(r.json()))
        if r.status_code == httpx.codes.OK:
            return True

    return False



from pathlib import Path
from typing import List, Optional, cast

import git
import typer

from phidata.infra.base import InfraConfig
from phidata.infra.docker.config import DockerConfig
from phidata.infra.k8s.config import K8sConfig
from phidata.infra.aws.config import AwsConfig
from phidata.workspace import WorkspaceConfig

from phi.backend_api.workspace_api import update_workspace, upsert_workspace
from phi.conf.constants import DEFAULT_WS_NAME
from phi.conf.phi_conf import PhiConf, PhiWsData
from phi.enums.user_enums import VersionControlProviderEnum
from phi.utils.cli_console import (
    print_conf_not_available_msg,
    print_error,
    print_heading,
    print_info,
    print_subheading,
    print_warning,
)
from phi.utils.common import is_empty, str_to_int
from phi.utils.dttm import current_datetime_utc, dttm_to_dttm_str
from phi.utils.filesystem import rmdir_recursive
from phi.utils.git import GitCloneProgress, get_remote_origin_for_dir
from phi.utils.log import logger
from phi.workspace.template_to_repo_map import template_to_repo_map
from phi.workspace.ws_enums import (
    WorkspaceEnv,
    WorkspaceInfra,
    WorkspaceStarterTemplate,
    WorkspaceVisibilityEnum,
)
from phi.workspace.ws_exceptions import WorkspaceConfigException
from phi.workspace.ws_schemas import (
    UpdateWorkspace,
    UpsertWorkspaceFromCli,
    WorkspaceSchema,
)
from phi.workspace.ws_utils import (
    get_ws_config_file_path,
    is_valid_ws_config_file_path,
    print_howtofix_pending_actions,
)

# List of reserved workspace names
RESERVED_WS_NAMES = ["root", "bedi", "ashpreet", "ashpreetbedi"]


def create_new_workspace(ws_name: str, template: WorkspaceStarterTemplate) -> bool:
    """Creates a new workspace for the user.
    This function clones a phidata template on the users machine at the path:
        cwd/ws_name
    cwd: current working dir - where the command was ran from.

    Steps:
    1. Validate ws_name and if we can create a folder with that name
    2. Clone the starter template in cwd/ws_name
    3. Delete the .git folder
    4. Provide the user with steps to add a remote origin for this workspace

    NOTE: till the user runs `phi ws setup` the workspace is not registered with phidata
        and can only be used locally
    """

    current_dir: Path = Path(".").resolve()

    # Phidata should be initialized before creating a workspace
    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return False

    # Error if the workspace name provided is a reserved name
    if ws_name in RESERVED_WS_NAMES:
        print_error(f"The name `{ws_name}` is reserved")
        return False

    # Check if a workspace with the same name exists
    _existing_ws_data: Optional[PhiWsData] = phi_conf.get_ws_data_by_name(ws_name)
    if _existing_ws_data is not None:
        print_error(f"Found existing record for a workspace: {ws_name}")
        print_info(
            f"\nNote: If you would like to setup the workspace instead, run `phi ws setup` from the workspace directory"
        )
        delete_existing_ws_data = typer.confirm(
            f"Would you like to delete existing workspace data and create a new workspace?",
            default=True,
        )
        if delete_existing_ws_data:
            phi_conf.delete_ws(ws_name)
        else:
            return False

    # Check if we can create the workspace in the current dir
    ws_dir_path: Path = current_dir.joinpath(ws_name)
    if ws_dir_path.exists():
        print_error(
            f"Directory {ws_dir_path} exists, please delete manually or choose another name for workspace"
        )
        return False

    print_info(f"Creating {str(ws_dir_path)}")
    # Clone the starter repo for this cloud in the new directory
    repo_to_clone = template_to_repo_map.get(template)
    logger.debug("Cloning: {}".format(repo_to_clone))
    try:
        _cloned_git_repo: git.Repo = git.Repo.clone_from(
            repo_to_clone, str(ws_dir_path), progress=GitCloneProgress()  # type: ignore
        )
    except Exception as e:
        print_error(e)
        return False

    # Remove existing .git folder
    _dot_git_folder = ws_dir_path.joinpath(".git")
    _dot_git_exists = _dot_git_folder.exists()
    if _dot_git_exists:
        # logger.debug(".git folder exists... deleting")
        _dot_git_exists = not rmdir_recursive(_dot_git_folder)
    # logger.debug(f"_dot_git_exists: {_dot_git_exists}")

    # validate that a workspace config file exists
    ws_config_file_path: Path = get_ws_config_file_path(ws_dir_path)
    # logger.debug(
    #     "Checking if workspace config file exists: {}".format(ws_config_file_path)
    # )
    if not (ws_config_file_path.exists() and ws_config_file_path.is_file()):
        print_error(f"Could not find workspace config at: {ws_config_file_path}")
        return False

    phi_conf.add_new_created_ws_to_config(
        ws_name=ws_name,
        ws_dir_path=ws_dir_path,
        ws_config_file_path=ws_config_file_path,
    )
    print_info(f"Your new workspace is available at {str(ws_dir_path)}\n")
    return setup_workspace(ws_dir_path=ws_dir_path)


def print_git_setup(
    ws_name: str,
    ws_dir_path: Path,
    version_control_provider: Optional[
        VersionControlProviderEnum
    ] = VersionControlProviderEnum.GITHUB,
) -> None:
    """Prints commands to setup a remote git repo for a workspace"""

    # Check if a remote origin is available
    _remote_origin_url: Optional[str] = get_remote_origin_for_dir(ws_dir_path)
    # logger.debug("_remote_origin_url: {}".format(_remote_origin_url))
    if _remote_origin_url:
        print_info(f"Remote origin already set: {_remote_origin_url}")
        return

    create_new_git_repo_url = "https://github.com/new"
    create_remote_repo_url: str
    # TODO: Add more repo urls
    if version_control_provider == VersionControlProviderEnum.GITHUB:
        create_remote_repo_url = create_new_git_repo_url
    else:
        create_remote_repo_url = create_new_git_repo_url

    print_subheading(f"\nSteps to setup a git repo for workspace: {ws_name}")
    print_info(f"\nCreate a new git repo named {ws_name} at {create_remote_repo_url}")
    print_info(
        f"Then follow the following steps to initialize the directory {ws_name} as a git repo and run git push"
    )
    print_info(
        f"NOTE: These steps will be provided by your version control provider (github, gitlab..), but would look something like:"
    )
    print_info(f"       git init")
    print_info(f"       git add .")
    print_info(f"       git commit -m 'Init Phidata Workspace'")
    print_info(f"       git branch -M main")
    print_info(
        f"       git remote add origin https://github.com/<USERNAME>/{ws_name}.git"
    )
    print_info(f"       git push -u origin main")
    print_info(f"Run `phi ws setup` after running git push to update your workspace\n")

    _open_browser = typer.confirm(
        f"Would you like to open your browser to visit: {create_remote_repo_url}\n-->"
    )
    if _open_browser:
        typer.launch(create_remote_repo_url)


def clone_workspace(ws_to_clone: WorkspaceSchema, phi_conf: PhiConf) -> None:
    """Clones an available phidata workspace on the users machine. To clone a ws, the PhiConf
    already has the ws_data and ws_schema available. We then update the
    ws_dir_path and ws_config_file_path in the ws_data. After updating the
    path, we're call setup_workspace() to go through our usual setup flow
    """

    # Check if we can create the workspace in the current dir
    current_dir: Path = Path("../../workspace").resolve()
    logger.debug(f"current_dir: {current_dir}")
    ws_dir_path: Path = current_dir.joinpath(ws_to_clone.name)
    if ws_dir_path.exists():
        print_error(
            f"{ws_dir_path} already exists, please choose another ws to clone or delete the existing dir and run `phi ws init` again"
        )
        return

    print_info(f"Creating {str(ws_dir_path)}")
    # Clone the workspace repo in ws_dir_path
    repo_to_clone = ws_to_clone.git_url
    _cloned_git_repo: git.Repo = git.Repo.clone_from(
        repo_to_clone, str(ws_dir_path), progress=GitCloneProgress()  # type: ignore
    )
    # Validate ws_dir_path exists and is a directory
    if not (ws_dir_path.exists() and ws_dir_path.is_dir()):
        print_error(f"Could not clone workspace, please try again")
        return
    # Get ws_config_file_path and validate
    ws_config_file_path: Path = get_ws_config_file_path(ws_dir_path)
    if not (ws_config_file_path.exists() and ws_config_file_path.is_file()):
        print_error(f"Could not find workspace config at: {ws_config_file_path}")
        return
    # Update the ws_data in PhiConf
    phi_conf.update_ws_data_for_cloned_ws(
        ws_name=ws_to_clone.name,
        ws_dir_path=ws_dir_path,
        ws_config_file_path=ws_config_file_path,
    )
    setup_workspace(ws_dir_path)


def create_workspace_using_input():
    """Takes input from the user and calls create_new_workspace()
    This function acts as an intermediary between initialize_workspace() and create_new_workspace()
    """

    # Get workspace name from user
    ws_name_inp_raw = input(f"Workspace name (default: {DEFAULT_WS_NAME}): ")
    if is_empty(ws_name_inp_raw):
        ws_name_inp_raw = DEFAULT_WS_NAME

    # Display available starter templates and ask user to select one
    print_info("Select Starter Template (default: docker)")
    templates = WorkspaceStarterTemplate.values_list()
    for idx, prvdr in enumerate(templates, start=1):
        print_info("  [{}] {}".format(idx, prvdr))

    # Get starter template from the user
    template_inp_raw = input("--> ")
    # Convert input to int value.
    template_inp = str_to_int(template_inp_raw)
    template: WorkspaceStarterTemplate = WorkspaceStarterTemplate.docker

    if template_inp is not None:
        try:
            template = cast(
                WorkspaceStarterTemplate,
                WorkspaceStarterTemplate.from_str(templates[template_inp - 1]),
            )
        except Exception as e:
            print_error(e)

    logger.debug("Selected: {}".format(template.value))
    print_info(
        f"Creating workspace {ws_name_inp_raw} using the {template.value} template\n"
    )
    create_new_workspace(ws_name_inp_raw, template)


def initialize_workspace() -> None:
    """Initialize a phidata workspace.
    This function is called by `phi ws init` to create/clone a new workspace.
    """
    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return

    # TODO: uncomment when backend is live
    # available_ws = phi_conf.available_ws
    # if available_ws:
    #     print_info("Would you like to:")
    #     print_info("  [1] Create a new workspace")
    #     print_info("  [2] Clone an existing workspace")
    #     print_info("  [3] Enter any other key to exit")
    #     ws_inp_raw = input("--> ")
    #     ws_inp_int = str_to_int(ws_inp_raw)
    #     if ws_inp_int == 1:
    #         create_workspace_using_input()
    #     elif ws_inp_int == 2:
    #         print_info("\nSelect workspace to clone:")
    #         for idx, avl_ws in enumerate(available_ws, start=1):
    #             print_info("  [{}] {}".format(idx, avl_ws.name))
    #         clone_ws_inp_raw = input("--> ")
    #         clone_ws_inp_int = str_to_int(clone_ws_inp_raw)
    #         if clone_ws_inp_int:
    #             ws_to_clone = available_ws[clone_ws_inp_int - 1]
    #             clone_workspace(ws_to_clone, phi_conf)
    #         return
    #     return
    # else:
    print_heading("Creating a new workspace")
    create_workspace_using_input()


def setup_workspace(ws_dir_path: Path) -> bool:
    """Set-up a phidata workspace at directory: `ws_dir_path`.
    This is the catchall function for a workspace. Run it in a directory and it
    should figure everything out.

    1. Validate pre-requisites
    1.1 Check ws_dir_path is valid
    1.2 Check PhiConf is valid
    1.3 Validate PhiWsData is available
        Get the ws_data using the ws_dir_path
        If ws_data is None, read the ws_config_file_path to map it to an available workspace
            This may happen if the user ran `phi init -r`
        If the user cloned this directory directly, there will be no record of this workspace in the PhiConf
        In that case, we manually add the ws to PhiConf and process the ws_config_file_path and setup this workspace.
    1.4 Check if this is the active workspace or if it needs to be set as active
    1.5 Check if remote origin is available

    2. Create or Update WorkspaceSchema
    If a ws_schema exists for this workspace, this workspace has been setup before
    2.1 Create WorkspaceSchema for a NEWLY CREATED WORKSPACE
        If ws_schema is None, this is a NEWLY CREATED WORKSPACE.
        i.e. A new workspace was created using create_new_workspace() <- which is called using `phi ws init`.
        This ws doesnt yet have a ws_schema, make an api call to zeus to create one.
        If the api call returns None, then we could not create the workspace and we exit
    2.2 Update WorkspaceSchema if remote origin has been updated
    2.3 Set as active if needed
    2.4 Refresh PhiWsData

    3. Complete Workspace setup
    3.1 Validate ws_config_file_path
    3.2 Validate PhiWsData
        Reaching here implies a all ws_data is available and updated.
        Validate the ws_data and print any pending setup actions.

    `phi ws setup` is a generic catch-all function. It should handle errors graciously
    and provide "how to fix" messages and "next steps" to get the user running.
    """

    print_heading("Running workspace setup\n")

    ######################################################
    ## 1. Validate Pre-requisites
    ######################################################
    ######################################################
    # 1.1 Check ws_dir_path is valid
    ######################################################
    _ws_is_valid: bool = (
        ws_dir_path is not None and ws_dir_path.exists() and ws_dir_path.is_dir()
    )
    if not _ws_is_valid:
        print_error("Invalid directory: {}".format(ws_dir_path))
        return False

    ######################################################
    # 1.2 Check PhiConf is valid
    ######################################################
    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return False

    ######################################################
    # 1.3 Validate PhiWsData is available
    ######################################################
    logger.debug(f"Checking for a workspace at {ws_dir_path}")
    ws_data: Optional[PhiWsData] = phi_conf.get_ws_data_by_path(ws_dir_path)
    if ws_data is None:
        # This happens if
        # - the user ran `phi init -r` which erases the self._path_to_ws_data_map
        # - OR the user is setting up a manual workspace
        logger.debug(f"Could not find a workspace at path: {ws_dir_path}")
        # We'll try to parse this workspace using the workspace config
        # validate that a workspace config file exists
        __ws_config_file_path: Path = get_ws_config_file_path(ws_dir_path)

        # In this case, the local workspace directory exists but PhiConf does not have a record
        print_info(f"Adding {ws_dir_path} as a workspace")
        phi_conf.add_new_created_ws_to_config(
            ws_name=ws_dir_path.stem,
            ws_dir_path=ws_dir_path,
            ws_config_file_path=__ws_config_file_path,
        )
        ws_data = phi_conf.get_ws_data_by_path(ws_dir_path)
        # If the ws_data is still none it means the workspace has bad data in it
        if ws_data is None:
            print_error(f"Could not create workspace")
            print_info(f"Workspace Dir     : {ws_dir_path}")
            print_info(f"Workspace Config   : {__ws_config_file_path}")
            print_error("Please try again")
            return False
    else:
        logger.debug(f"Found workspace {ws_data.ws_name}")
        phi_conf.refresh_ws_config(ws_data.ws_name)

    ######################################################
    # 1.4 Set workspace as active if needed
    ######################################################
    set_as_active_ws = False

    # If this is the users first workspace i.e. No active_ws is available, then set as active_ws
    if phi_conf.active_ws_name is None:
        set_as_active_ws = True
    # otherwise, ask the user if this should be the active ws
    elif phi_conf.active_ws_name != ws_data.ws_name:
        _set_as_active_ws_inp = typer.confirm(
            f"Would you like to set this as your active workspace",
            default=True,
        )
        if _set_as_active_ws_inp:
            set_as_active_ws = True

    if set_as_active_ws:
        phi_conf.active_ws_name = ws_data.ws_name

    ######################################################
    # 1.5 Check if remote origin is available
    ######################################################
    _remote_origin_url: Optional[str] = get_remote_origin_for_dir(ws_dir_path)
    logger.debug("Remote_origin_url: {}".format(_remote_origin_url))

    # TODO: uncomment this block when the backend is up and running
    ######################################################
    ## 2. Create or Update WorkspaceSchema
    ######################################################
    # If a ws_schema exists for this workspace, this workspace has been setup before
    # ws_schema: Optional[WorkspaceSchema] = ws_data.ws_schema

    ######################################################
    # 2.1 Create WorkspaceSchema for a NEWLY CREATED WORKSPACE
    ######################################################
    # if ws_schema is None:
    #     # If ws_schema is None, this is a NEWLY CREATED WORKSPACE.
    #     # We make a call to the backend to create a new ws_schema
    #     logger.debug("ws_schema not found, trying to create one... ")
    #     logger.debug("ws_name: {}".format(ws_data.ws_name))
    #     logger.debug("set_as_active_ws: {}".format(_is_active_ws))
    #
    #     ws_schema = upsert_workspace(
    #         UpsertWorkspaceFromCli(
    #             name=ws_data.ws_name,
    #             visibility=WorkspaceVisibilityEnum.PRIVATE.value,
    #             git_url=_remote_origin_url,
    #             setup_complete=True,
    #             extra_data={"upserted_at": dttm_to_dttm_str(current_datetime_utc())},
    #             is_active_ws_for_user=_is_active_ws,
    #             only_create=True,
    #         )
    #     )
    #     if ws_schema is not None:
    #         phi_conf.update_ws_schema_for_workspace(
    #             ws_schema=ws_schema, set_as_active=_is_active_ws
    #         )

    ######################################################
    # 2.2 Update WorkspaceSchema if available
    ######################################################
    # if ws_schema is not None:
    #     if _remote_origin_url is not None and _remote_origin_url != ws_schema.git_url:
    #         print_info(f"Updating workspace git origin to: {_remote_origin_url}")
    #         ws_schema_updated = update_workspace(
    #             UpdateWorkspace(
    #                 id_workspace=ws_schema.id_workspace,
    #                 git_url=_remote_origin_url,
    #             )
    #         )
    #         if ws_schema_updated is not None:
    #             # Update the ws_schema for this workspace.
    #             phi_conf.update_ws_data(
    #                 ws_name=ws_schema.name,
    #                 ws_schema=ws_schema_updated,
    #             )
    #         else:
    #             print_error(
    #                 "Could not update git origin, please run `phi ws setup` again"
    #             )
    #         logger.debug(f"WorkspaceSchema git url updated to: {ws_schema.git_url}")

    ######################################################
    # 2.3 Refresh PhiWsData
    ######################################################
    # Refresh ws_data because phi_conf.update_ws_schema_for_workspace() or phi_conf.update_ws_data()
    # will create a new ws_data object in PhiConf
    # Also, cast it to PhiWsData because ws_data is no longer Optional[PhiWsData]
    ws_data = cast(PhiWsData, phi_conf.get_ws_data_by_name(ws_data.ws_name))

    ######################################################
    ## 3. Complete Workspace setup
    ######################################################

    ######################################################
    # 3.1 Validate ws_config_file_path
    ######################################################
    _ws_config_file_path: Optional[Path] = ws_data.ws_config_file_path
    if not is_valid_ws_config_file_path(_ws_config_file_path):
        print_error("Could not validate workspace config")
        return False

    ######################################################
    # 3.2 Validate PhiWsData
    ######################################################
    ws_is_valid, pending_actions = phi_conf.validate_workspace(ws_name=ws_data.ws_name)

    if ws_is_valid:
        scripts_dir = ws_data.ws_config.scripts_dir
        install_deps_file = f"source {ws_dir_path}/{scripts_dir}/install_ws.sh"
        print_heading(f"Setup complete! Next steps:\n")
        print_info("1. Update your workspace config:")
        print_info(f"\t{str(_ws_config_file_path)}")
        print_info("2. Install your workspace:")
        print_info(f"\t{install_deps_file}")
        print_info("3. Dry-run workspace deploy:")
        print_info("\tphi ws up -dr")
        print_info("4. Deploy your workspace:")
        print_info("\tphi ws up")
        # if pending_actions and len(pending_actions) > 0:
        #     print_info("3. Complete pending actions:")
        #     print_howtofix_pending_actions(pending_actions)
        return True
    else:
        print_info(f"Workspace setup pending.")
        print_howtofix_pending_actions(pending_actions)
        return False

    ######################################################
    ## End Workspace setup
    ######################################################


def deploy_workspace(
    ws_data: PhiWsData,
    target_env: Optional[WorkspaceEnv] = None,
    target_infra: Optional[WorkspaceInfra] = None,
    target_name: Optional[str] = None,
    target_type: Optional[str] = None,
    dry_run: Optional[bool] = False,
) -> None:
    """Deploy a Phidata Workspace. This is called from `phi ws up`"""

    if ws_data is None or ws_data.ws_config is None:
        print_error("WorkspaceConfig invalid")
        return
    ws_config: WorkspaceConfig = ws_data.ws_config

    # Add configs in the order they should be deployed
    configs_to_deploy_infra_filter: List[InfraConfig] = []
    if target_infra is None:
        if ws_config.docker is not None:
            configs_to_deploy_infra_filter.extend(ws_config.docker)
        if ws_config.aws is not None:
            configs_to_deploy_infra_filter.extend(ws_config.aws)
        if ws_config.k8s is not None:
            configs_to_deploy_infra_filter.extend(ws_config.k8s)
    elif target_infra == WorkspaceInfra.docker:
        if ws_config.docker is not None:
            configs_to_deploy_infra_filter.extend(ws_config.docker)
        else:
            print_error("No DockerConfig provided")
    elif target_infra == WorkspaceInfra.k8s:
        if ws_config.k8s is not None:
            configs_to_deploy_infra_filter.extend(ws_config.k8s)
        else:
            print_error("No K8sConfig provided")
    elif target_infra == WorkspaceInfra.aws:
        if ws_config.aws is not None:
            configs_to_deploy_infra_filter.extend(ws_config.aws)
        else:
            print_error("No AwsConfig provided")

    configs_to_deploy_env_filter: List[InfraConfig] = []
    if target_env is None:
        configs_to_deploy_env_filter = configs_to_deploy_infra_filter
    else:
        for config in configs_to_deploy_infra_filter:
            if config.env is None:
                continue
            if config.env.lower() not in WorkspaceEnv.values_list():
                print_error(f"{config.env} not supported")
                continue
            config_env = WorkspaceEnv.from_str(config.env)

            if config_env == target_env:
                configs_to_deploy_env_filter.append(config)
    # logger.debug("Deploying configs: {}".format(configs_to_deploy_env_filter))

    num_configs_to_process = len(configs_to_deploy_env_filter)
    num_configs_processed = 0
    for config in configs_to_deploy_env_filter:
        if not isinstance(config, InfraConfig):
            print_error(f"{config} is not an instance of InfraConfig")
            continue
        if not config.is_valid():
            print_error("Config invalid")
            continue
        if not config.enabled:
            logger.debug(
                f"{config.__class__.__name__} for env {config.__class__.env} disabled"
            )
            num_configs_processed += 1
            continue

        ######################################################################
        # NOTE: VERY IMPORTANT TO GET RIGHT
        # Update sub-configs data using WorkspaceConfig
        # 1. Pass down the paths from the WorkspaceConfig
        #       These paths are used everywhere from Infra to Apps
        # 2. Pass down local_env, docker_env, k8s_env for building PhidataApps
        # 3. Pass down common cloud configuration. eg: aws_region, aws_profile
        ######################################################################

        # The ws_dir_path is the ROOT directory for the workspace
        config.workspace_root_path = ws_data.ws_dir_path
        config.workspace_config_file_path = ws_data.ws_config_file_path
        config.scripts_dir = ws_config.scripts_dir
        config.storage_dir = ws_config.storage_dir
        config.meta_dir = ws_config.meta_dir
        config.products_dir = ws_config.products_dir
        config.notebooks_dir = ws_config.notebooks_dir
        config.workspace_config_dir = ws_config.workspace_config_dir

        print_info("")
        logger.debug(f"Deploying {config.__class__.__name__}")
        if isinstance(config, DockerConfig):
            from phi.docker.docker_operator import deploy_docker_config

            # Pass down docker_env from WorkspaceConfig
            config.docker_env = ws_config.docker_env
            deploy_docker_config(
                config=config,
                name_filter=target_name,
                type_filter=target_type,
                dry_run=dry_run,
            )
            num_configs_processed += 1
        if isinstance(config, K8sConfig):
            from phi.k8s.k8s_operator import deploy_k8s_config

            # Pass down k8s_env from WorkspaceConfig
            config.k8s_env = ws_config.k8s_env
            deploy_k8s_config(
                config=config,
                name_filter=target_name,
                type_filter=target_type,
                dry_run=dry_run,
            )
            num_configs_processed += 1
        if isinstance(config, AwsConfig):
            from phi.aws.aws_operator import deploy_aws_config

            # Pass down common aws configuration from WorkspaceConfig
            # logger.debug("Updating AwsConfig from WorkspaceConfig")
            config.aws_region = ws_config.aws_region
            config.aws_profile = ws_config.aws_profile
            config.aws_config_file = ws_config.aws_config_file
            config.aws_shared_credentials_file = ws_config.aws_shared_credentials_file
            deploy_aws_config(
                config=config,
                name_filter=target_name,
                type_filter=target_type,
                dry_run=dry_run,
            )
            num_configs_processed += 1

    print_info(
        f"\n# Configs processed: {num_configs_processed}/{num_configs_to_process}"
    )
    if num_configs_to_process == num_configs_processed:
        if not dry_run:
            print_subheading("Workspace deploy success")
    else:
        print_error("Workspace deploy failed")


def shutdown_workspace(
    ws_data: PhiWsData,
    target_env: Optional[WorkspaceEnv] = None,
    target_infra: Optional[WorkspaceInfra] = None,
    target_name: Optional[str] = None,
    target_type: Optional[str] = None,
    dry_run: Optional[bool] = False,
) -> None:
    """Shutdown the Phidata Workspace. This is called from `phi ws down`"""

    if ws_data is None or ws_data.ws_config is None:
        print_error("WorkspaceConfig invalid")
        return

    ws_config: WorkspaceConfig = ws_data.ws_config

    # Add configs in the order they should be shut down
    configs_to_shutdown_infra_filter: List[InfraConfig] = []
    if target_infra is None:
        if ws_config.docker is not None:
            configs_to_shutdown_infra_filter.extend(ws_config.docker)
        if ws_config.k8s is not None:
            configs_to_shutdown_infra_filter.extend(ws_config.k8s)
        if ws_config.aws is not None:
            configs_to_shutdown_infra_filter.extend(ws_config.aws)
    elif target_infra == WorkspaceInfra.docker:
        if ws_config.docker is not None:
            configs_to_shutdown_infra_filter.extend(ws_config.docker)
        else:
            print_error("No DockerConfig provided")
    elif target_infra == WorkspaceInfra.k8s:
        if ws_config.k8s is not None:
            configs_to_shutdown_infra_filter.extend(ws_config.k8s)
        else:
            print_error("No K8sConfig provided")
    elif target_infra == WorkspaceInfra.aws:
        if ws_config.aws is not None:
            configs_to_shutdown_infra_filter.extend(ws_config.aws)
        else:
            print_error("No AwsConfig provided")

    configs_to_shutdown_env_filter: List[InfraConfig] = []
    if target_env is None:
        configs_to_shutdown_env_filter = configs_to_shutdown_infra_filter
    else:
        for config in configs_to_shutdown_infra_filter:
            if config.env is None:
                continue
            if config.env.lower() not in WorkspaceEnv.values_list():
                print_error(f"{config.env} not supported")
                continue
            config_env = WorkspaceEnv.from_str(config.env)

            if config_env == target_env:
                configs_to_shutdown_env_filter.append(config)
    # logger.debug("Shutting down configs: {}".format(configs_to_shutdown_env_filter))

    num_configs_to_process = len(configs_to_shutdown_env_filter)
    num_configs_processed = 0
    for config in configs_to_shutdown_env_filter:
        if not isinstance(config, InfraConfig):
            print_error(f"{config} is not an instance of InfraConfig")
            continue
        if not config.is_valid():
            print_error("Config invalid")
            continue
        if not config.enabled:
            logger.debug(
                f"{config.__class__.__name__} for env {config.__class__.env} disabled"
            )
            num_configs_processed += 1
            continue

        ######################################################################
        # NOTE: VERY IMPORTANT TO GET RIGHT
        # Update sub-configs data using WorkspaceConfig
        # 1. Pass down the paths from the WorkspaceConfig
        #       These paths are used everywhere from Infra to Apps
        # 2. Pass down common cloud configuration. eg: aws_region, aws_profile
        ######################################################################

        # The ws_dir_path is the ROOT directory for the workspace
        config.workspace_root_path = ws_data.ws_dir_path
        config.workspace_config_file_path = ws_data.ws_config_file_path
        config.scripts_dir = ws_config.scripts_dir
        config.storage_dir = ws_config.storage_dir
        config.meta_dir = ws_config.meta_dir
        config.products_dir = ws_config.products_dir
        config.notebooks_dir = ws_config.notebooks_dir
        config.workspace_config_dir = ws_config.workspace_config_dir

        print_info("")
        logger.debug(f"Shutting down {config.__class__.__name__}")
        if isinstance(config, DockerConfig):
            from phi.docker.docker_operator import shutdown_docker_config

            # Pass down docker_env from WorkspaceConfig
            config.docker_env = ws_config.docker_env
            shutdown_docker_config(
                config=config,
                name_filter=target_name,
                type_filter=target_type,
                dry_run=dry_run,
            )
            num_configs_processed += 1
        if isinstance(config, K8sConfig):
            from phi.k8s.k8s_operator import shutdown_k8s_config

            # Pass down k8s_env from WorkspaceConfig
            config.k8s_env = ws_config.k8s_env
            shutdown_k8s_config(
                config=config,
                name_filter=target_name,
                type_filter=target_type,
                dry_run=dry_run,
            )
            num_configs_processed += 1
        if isinstance(config, AwsConfig):
            from phi.aws.aws_operator import shutdown_aws_config

            # Pass down common aws configuration from WorkspaceConfig
            config.aws_region = ws_config.aws_region
            config.aws_profile = ws_config.aws_profile
            config.aws_config_file = ws_config.aws_config_file
            config.aws_shared_credentials_file = ws_config.aws_shared_credentials_file
            shutdown_aws_config(
                config=config,
                name_filter=target_name,
                type_filter=target_type,
                dry_run=dry_run,
            )
            num_configs_processed += 1

    print_info(
        f"\n# Configs processed: {num_configs_processed}/{num_configs_to_process}"
    )
    if num_configs_to_process == num_configs_processed:
        if not dry_run:
            print_subheading("Workspace shutdown success")
    else:
        print_subheading("Workspace shutdown failed")


def init_workspace_data(ws_schema: WorkspaceSchema, phi_conf: PhiConf):
    """Initializes workspace data like cloud projects"""

    logger.debug("*=* Initializing workspace data for: {}".format(ws_schema.name))
    logger.debug("TO_BE_IMPLEMENTED")
    logger.debug("*=* Workspace data init complete: {}".format(ws_schema.name))


def set_workspace_as_active(ws_name: Optional[str], refresh: bool = True) -> bool:

    ######################################################
    ## 1. Validate Pre-requisites
    ######################################################
    ######################################################
    # 1.1 Check PhiConf is valid
    ######################################################
    phi_conf: Optional[PhiConf] = PhiConf.get_saved_conf()
    if not phi_conf:
        print_conf_not_available_msg()
        return False

    ######################################################
    # 1.2 Check ws_dir_path is valid
    ######################################################
    # By default, we assume this command is run from the workspace directory
    ws_dir_path: Optional[Path] = None
    if ws_name is None:
        # If the user does not provide a ws_name, that implies `phi set` is ran from
        # the workspace directory.
        ws_dir_path = Path(".").resolve()
    else:
        # If the user provides a workspace name manually, we find the dir for that ws
        ws_dir_path = phi_conf.get_ws_dir_path_by_name(ws_name)
        if ws_dir_path is None:
            print_error(f"Could not find workspace {ws_name}")
            return False

    ws_dir_is_valid: bool = (
        ws_dir_path is not None and ws_dir_path.exists() and ws_dir_path.is_dir()
    )
    if not ws_dir_is_valid:
        print_error("Invalid workspace directory: {}".format(ws_dir_path))
        return False

    ######################################################
    # 1.3 Validate PhiWsData is available i.e. a workspace is available at this directory
    ######################################################
    logger.debug(f"Checking for a workspace at path: {ws_dir_path}")
    active_ws_data: Optional[PhiWsData] = phi_conf.get_ws_data_by_path(ws_dir_path)
    if active_ws_data is None:
        # This happens when the workspace is not yet setup
        print_info(f"Could not find a workspace at path: {ws_dir_path}")
        print_info(
            f"If this workspace has not been setup, please run `phi ws setup` from the workspace directory"
        )
        return False

    new_active_ws_name: str = active_ws_data.ws_name
    print_heading(f"Setting workspace {new_active_ws_name} as active")
    if refresh:
        try:
            refresh_success = phi_conf.refresh_ws_config(new_active_ws_name)
            if not refresh_success:
                print_warning("Could not refresh workspace config")
        except WorkspaceConfigException as e:
            print_info("Please fix the errors and try again")
            return False

    ######################################################
    # 1.4 Check if this is already the active workspace
    ######################################################
    current_active_ws_name = phi_conf.active_ws_name
    if new_active_ws_name != current_active_ws_name:
        # If a ws_schema exists for this workspace, this workspace has been authenticated
        # with the backend. In this case we need to send an API request to update
        # the active workspace for this user
        ws_schema: Optional[WorkspaceSchema] = active_ws_data.ws_schema
        if ws_schema is not None:
            print_info("Updating active workspace with the backend")
            # If ws_schema is None, this is a local workspace

    ######################################################
    ## 2. Set workspace as active
    ######################################################
    phi_conf.active_ws_name = new_active_ws_name
    print_info("Active workspace updated")
    return True



# import json
# from pathlib import Path
# from typing import Any, Dict, Optional
#
# import typer
#
# from pak8.conf import Pak8Conf
# from phi import schemas
# from phi.conf.phi_conf import PhiConf, PhiGcpData, PhiWsData
# from phi.utils.common import (
#     pprint_info,
#     pprint_error,
#     pprint_heading,
#     pprint_info,
#     pprint_status,
#     pprint_subheading,
#     str_to_int,
# )
# from phi.utils.filesystem import delete_from_fs
# from phi.utils.log import logger
# from phi.zeus_api.gcp import (
#     get_gke_cluster_for_gcp_project,
#     get_primary_gcp_project,
#     upsert_gcp_project,
# )
#
#
# def get_default_sa_key_path_(ws_dir_path: Path) -> Path:
#     return ws_dir_path.joinpath("phi").joinpath("keys").joinpath("phidata-sa-pk.json")
#
#
# def _upsert_primary_gcp_project_for_ws(
#     ws_name: str, config: PhiConf
# ) -> Optional[schemas.GCPProjectSchema]:
#     """For a given ws, reads the current GCP data and makes a call to zeus to create or update
#     the schemas.GCPProjectSchema for this workspace. We also set this project as primary project for this workspace.
#
#     Note: We use ws_name to get existing ws_gcp_data because prior to calling this function
#     the authenticate_gcp_project_for_ws() function would update the ws_gcp_data with the
#     service_account and key.
#     """
#     ws_schema: Optional[schemas.WorkspaceSchema] = config.get_ws_schema_by_name(ws_name)
#     ws_gcp_data: Optional[PhiGcpData] = config.get_ws_gcp_data_by_name(ws_name)
#     ws_pak8_conf: Optional[Pak8Conf] = config.get_ws_pak8_conf_by_name(ws_name)
#     if ws_gcp_data is None or ws_schema is None:
#         pprint_error(
#             "Could not authenticate GCP Project, please run `phi gcp auth` again"
#         )
#         return None
#
#     gcp_project: Optional[schemas.GCPProjectSchema] = ws_gcp_data.gcp_project
#     gcp_project_id = ws_gcp_data.gcp_project_id
#     logger.debug(f"Upserting GCPProjectSchema for {ws_name}: {gcp_project_id}")
#     gcp_project_schema_from_api: Optional[
#         schemas.GCPProjectSchema
#     ] = upsert_gcp_project(
#         schemas.UpsertGCPProjectFromCli(
#             name=gcp_project_id,
#             id_project=f"{ws_schema.id_workspace}_{gcp_project_id}",
#             id_workspace=ws_schema.id_workspace,
#             gcp_project_id=gcp_project_id,
#             gcp_service_account_key=ws_gcp_data.gcp_svc_account_key,
#             gcp_service_account=ws_gcp_data.gcp_svc_account,
#             gcp_zone=ws_pak8_conf.gcp.zone
#             if ws_pak8_conf and ws_pak8_conf.gcp
#             else None,
#             vpc_network=ws_pak8_conf.gcp.vpc_network
#             if ws_pak8_conf and ws_pak8_conf.gcp
#             else None,
#             set_as_primary_project_for_ws=True,
#         )
#     )
#     if gcp_project_schema_from_api:
#         config.update_ws_gcp_data(
#             ws_name=ws_schema.name, gcp_project=gcp_project_schema_from_api
#         )
#         gcp_project = ws_gcp_data.gcp_project
#         return gcp_project
#     # This codeblock would be used for bypassing zeus in creating a gcp project
#     # else:
#     #     gcp_project_schema: schemas.GCPProjectSchema = schemas.GCPProjectSchema(
#     #         name=gcp_project_id,
#     #         is_valid=True,
#     #         is_verified=False,
#     #         is_test=False,
#     #         gcp_project_id=gcp_project_id,
#     #         id_project=f"{ws_schema.id_workspace}_{gcp_project_id}",
#     #         id_workspace=ws_schema.id_workspace,
#     #         created_by_id_user=user.id_user,
#     #         created_by_email=user.email,
#     #         create_ts=current_datetime_utc(),
#     #         gcp_zone=ws_pak8_conf.gcp.zone if ws_pak8_conf and ws_pak8_conf.gcp else None,
#     #         gcp_service_account_key_available=True if ws_gcp_data.gcp_svc_account_key is not None else False,
#     #         vpc_network=ws_pak8_conf.gcp.vpc_network if ws_pak8_conf and ws_pak8_conf.gcp else None,
#     #         last_update_ts=None,
#     #         project_status=enums.GCPProjectStatus.NEW,
#     #     )
#
#     return gcp_project
#
#
# def create_service_account_key_application_default_creds(
#     ws_name: str,
#     config: PhiConf,
#     phi_sa_name: str,
#     phi_sa_displayname: str,
#     sa_key_path: Path,
# ) -> Optional[Dict[str, Any]]:
#     """Creates a GCP Service Account and returns the Service Account JSON Key using the
#     Google API Client Library: https://github.com/googleapis/google-zeus_api-python-client
#
#     Google has 2 types of client libraries: https://cloud.google.com/apis/docs/client-libraries-explained
#     1. Cloud Client Libraries (preferred)
#     2. Api Client Libraries (in maintenance mode now)
#     Ideally, we would use the IAM Cloud Client Library: https://github.com/googleapis/python-iam
#     But that doesn't have the ability to create service accounts yet, issue: https://github.com/googleapis/python-iam/issues/10
#     So we use Api Client Library with IAM V1 Service: http://googleapis.github.io/google-zeus_api-python-client/docs/dyn/iam_v1.html
#
#     Reference:
#     * https://cloud.google.com/iam/docs/creating-managing-service-accounts#iam-service-accounts-create-python
#     * https://cloud.google.com/iam/docs/reference/rest/v1/projects.serviceAccounts.keys
#     """
#
#     ######################################################
#     # To perform any kind of GCP actions, we need a "Credentials" object of type: google.auth.credentials.Credentials
#     # First we check if the Application Default Credentials are available
#     # https://googleapis.dev/python/google-auth/latest/user-guide.html#application-default-credentials
#     #
#     # For now ADCs are working and hopefully they will continue to work.
#     # If these stop working, we may have to revert to creating credentials
#     # using OAuth 2.0 for Installed Applications with the `google-auth-oauthlib` library
#     # https://github.com/googleapis/google-zeus_api-python-client/blob/master/docs/oauth-installed.md
#     ######################################################
#     import google.auth  # type: ignore
#     import google.oauth2  # type: ignore
#     import googleapiclient.discovery  # type: ignore
#
#     # The Service Account Key which will be used by phidata to manage GCP resources for the user
#     sa_key: Optional[Dict[str, Any]] = None
#
#     ######################################################
#     # Check if GCP Application Default Credentials are avialable
#     ######################################################
#     gcp_default_creds: Optional[google.oauth2.credentials.Credentials] = None
#     gcp_default_project: Optional[str] = None
#     try:
#         # google.auth.default() function docstring lists return type: Tuple[~google.auth.credentials.Credentials, Optional[str]]
#         # But upon inspecting the object, the credentials object returned is google.oauth2.credentials.Credentials
#         # which is a subclass of google.auth.credentials.Credentials
#         gcp_default_creds, gcp_default_project = google.auth.default(
#             scopes=["https://www.googleapis.com/auth/cloud-platform"]
#         )
#         # logger.debug("project: {}".format(gcp_default_project))
#         # logger.debug("credentials: {}".format(gcp_default_creds))
#         # logger.debug("refresh_token: {}".format(gcp_default_creds.refresh_token))
#         # logger.debug("client_id: {}".format(gcp_default_creds.client_id))
#         # logger.debug("valid: {}".format(gcp_default_creds.valid))
#     except google.auth.exceptions.DefaultCredentialsError as e:
#         logger.exception(e)
#
#     if gcp_default_creds is None:
#         pprint_error(
#             "gcloud default credentials unavailable, please authenticate the gcloud sdk and try again."
#         )
#         pprint_info("")
#         pprint_info("\tgcloud auth application-default login")
#         pprint_info("")
#         pprint_info(
#             "If you do not have the gcloud sdk installed, please follow the gcloud docs: https://cloud.google.com/sdk/docs"
#         )
#         return sa_key
#     if gcp_default_project is None:
#         pprint_error("No GCP Project available, please set one using:")
#         pprint_info("")
#         pprint_info("\tgcloud config set project <YOUR_GCP_PROJECT_HERE>")
#         pprint_info("")
#         return sa_key
#
#     # since this could be the first time we're creating a PhiGcpData for this ws
#     # set the gcp_project_id
#     # We have also confirmed that ADCs are available, so we set the flag to True as well
#     config.update_ws_gcp_data(
#         ws_name=ws_name,
#         gcp_project_id=gcp_default_project,
#         gcloud_default_creds_avl=True,
#     )
#     pprint_info("")
#     pprint_status(
#         f"Using gcloud default credentials for project: {gcp_default_project} to create a service account {phi_sa_displayname}"
#     )
#
#     ######################################################
#     # Build iAM Service to use to create the sa and key
#     ######################################################
#     iam_v1_service: Optional[googleapiclient.discovery.Resource] = None
#     try:
#         iam_v1_service = googleapiclient.discovery.build(
#             "iam", "v1", credentials=gcp_default_creds
#         )
#     except Exception as e:
#         logger.debug("could not create iam_v1_service")
#         logger.exception(e)
#     if iam_v1_service is None:
#         pprint_error("Could not connect to the Google IAM Service, please try again")
#         return sa_key
#
#     # Get all Service Accounts to first check if phidata already has a service account
#     _phi_svc_account: Optional[Dict[str, Any]] = None
#     # Follow: https://cloud.google.com/iam/docs/creating-managing-service-accounts#iam-service-accounts-create-python
#     _list_svc_accounts_resp: Optional[Dict[str, Any]] = (
#         iam_v1_service.projects()
#         .serviceAccounts()
#         .list(name=f"projects/{gcp_default_project}")
#         .execute()
#     )
#     if (
#         isinstance(_list_svc_accounts_resp, dict)
#         and "accounts" in _list_svc_accounts_resp
#     ):
#         # logger.debug("_list_svc_accounvts_resp: {}".format(_list_svc_accounts_resp))
#         # logger.debug("_list_svc_accounts_resp type: {}".format(type(_list_svc_accounts_resp)))
#         for _svc_accnt in _list_svc_accounts_resp["accounts"]:
#             if isinstance(_svc_accnt, dict):
#                 if _svc_accnt.get("displayName", None) == phi_sa_displayname:
#                     _phi_svc_account = _svc_accnt
#                     # break
#                 # logger.debug(f"svc accnt type: {type(_svc_accnt)}")
#                 # logger.debug(f"Found Service Account: {_svc_accnt}")
#
#     if _phi_svc_account is not None:
#         pprint_status("Found existing Service Account for phidata")
#         # logger.debug(f"_phi_svc_account type: {type(_phi_svc_account)}")
#         # logger.debug(f"{_phi_svc_account}")
#     else:
#         pprint_status("Creating a new Service Account for phidata")
#         _new_svc_account = (
#             iam_v1_service.projects()
#             .serviceAccounts()
#             .create(
#                 name=f"projects/{gcp_default_project}",
#                 body={
#                     "accountId": phi_sa_name,
#                     "serviceAccount": {"displayName": phi_sa_displayname},
#                 },
#             )
#             .execute()
#         )
#         # logger.debug("Created service account: {}".format(_new_svc_account))
#         # TODO: validate _new_svc_account before using
#         _phi_svc_account = _new_svc_account
#
#     if _phi_svc_account is None:
#         pprint_error(
#             "Could not create a service account for phidata, please run `phi gcp auth` again"
#         )
#         return sa_key
#
#     # Now that a service account is available, update the PhiGcpData for this ws
#     config.update_ws_gcp_data(
#         ws_name=ws_name,
#         gcp_svc_account=_phi_svc_account,
#     )
#
#     ######################################################
#     # Create the Service Account Key if needed
#     ######################################################
#     pprint_status("Creating a Service Account Key for phidata")
#     # Get the service account email
#     _phi_sa_email = _phi_svc_account["email"]
#     # https://cloud.google.com/iam/docs/creating-managing-service-account-keys
#     try:
#         _sa_key_resp = (
#             iam_v1_service.projects()
#             .serviceAccounts()
#             .keys()
#             .create(
#                 name=f"projects/{gcp_default_project}/serviceAccounts/{_phi_sa_email}"
#                 # TODO: Possibly add body={'privateKeyType':'TYPE_GOOGLE_CREDENTIALS_FILE'} here
#             )
#             .execute()
#         )
#     except googleapiclient.errors.HttpError as e:
#         _reason = e._get_reason()
#         pprint_error("Error while creating service account key")
#         pprint_error(_reason)
#         return sa_key
#
#     # logger.debug("_sa_key_resp: {}".format(_sa_key_resp))
#     # logger.debug("_sa_key_resp type: {}".format(type(_sa_key_resp)))
#
#     # Validate the key we just created before using it
#     _sa_key_private_key_type: Optional[str] = _sa_key_resp.get("privateKeyType", None)
#     if (
#         _sa_key_private_key_type is not None
#         and _sa_key_private_key_type != "TYPE_GOOGLE_CREDENTIALS_FILE"
#     ):
#         pprint_error("Error while creating service account key")
#         pprint_error(
#             f"Unexpected `privateKeyType` in Service Account Key, found {_sa_key_private_key_type} expected 'TYPE_GOOGLE_CREDENTIALS_FILE'"
#         )
#         return sa_key
#
#     sa_key = _sa_key_resp
#     if sa_key is None:
#         pprint_error(
#             "Could not create a service account key for phidata, please run `phi gcp auth` again"
#         )
#         return sa_key
#
#     # Set the service account key in the PhiGcpData
#     # Then save it in a file on the users machine
#     config.update_ws_gcp_data(
#         ws_name=ws_name,
#         gcp_svc_account_key=sa_key,
#     )
#     pprint_info(
#         f"Saving Service Account Key to file {str(sa_key_path)}\nPlease ensure that you don't check the key to version control"
#     )
#     if sa_key_path.exists():
#         pprint_info("")
#         pprint_status(f"Deleting existing file at {str(sa_key_path)}")
#         delete_from_fs(sa_key_path)
#
#     try:
#         sa_key_path.parent.mkdir(parents=True, exist_ok=True)
#         sa_key_path.write_text(json.dumps(sa_key))
#     except Exception as e:
#         logger.exception(e)
#         pprint_error(
#             f"Could not write service account key to file: {str(sa_key_path)}\nplease run `phi gcp auth` again"
#         )
#         return None
#
#     try:
#         _list_sa_keys_resp = (
#             iam_v1_service.projects()
#             .serviceAccounts()
#             .keys()
#             .list(
#                 name=f"projects/{gcp_default_project}/serviceAccounts/{_phi_sa_email}"
#             )
#             .execute()
#         )
#         # logger.debug("_list_sa_keys_resp: {}".format(_list_sa_keys_resp))
#         # logger.debug("_list_sa_keys_resp type: {}".format(type(_list_sa_keys_resp)))
#         _available_keys = _list_sa_keys_resp.get("keys", [])
#         _num_avl_keys = len(_available_keys)
#         if _num_avl_keys > 0:
#             pprint_status(f"This service account has {_num_avl_keys} active keys")
#     except Exception as e:
#         pass
#
#     return sa_key
#
#
# def create_service_account_key_manually(
#     ws_name: str,
#     config: PhiConf,
#     phi_sa_name: str,
#     phi_sa_displayname: str,
#     sa_key_path: Path,
#     ws_config_file_path: Path,
# ) -> Optional[Dict[str, Any]]:
#
#     # The Service Account Key which will be used by phidata to manage GCP resources for the user
#     sa_key: Optional[Dict[str, Any]] = None
#
#     pprint_subheading("Steps to create a service account and key:")
#     # PREREQUISITES
#     pprint_info(
#         "Open another terminal and create a service account using the following steps:"
#     )
#     pprint_subheading("\nI. Setup GCloud SDK (skip if you already have gcloud setup)")
#     pprint_info("  1. Install gcloud: https://cloud.google.com/sdk/docs")
#     pprint_info("  2. Initialize gcloud using `gcloud init`")
#     pprint_info("  3. Please select the GCP project for phidata to use")
#
#     pprint_subheading("\nII. Create Service Account for phidata and grant permissions")
#     pprint_info(
#         f"  1. Create a service account using:\n\ngcloud iam service-accounts create {phi_sa_name} --display-name='{phi_sa_displayname}'\n"
#     )
#     pprint_info(
#         f"  2. Get the service account email using:\n\ngcloud iam service-accounts list --filter='NAME:{phi_sa_displayname}' --format='value(EMAIL)'\n"
#     )
#     pprint_info(
#         f"  3. Grant the Service Account permissions to manage resources on your behalf\n"
#     )
#     pprint_info(f"gcloud projects add-iam-policy-binding [PROJECT_ID] \ ")
#     pprint_info(f"--member='serviceAccount:[SERVICE_ACCOUNT_EMAIL]' \ ")
#     pprint_info(f"--role='roles/editor'\n")
#     pprint_info(f"  4. Finally, create a key so phi can use the service account\n")
#     pprint_info(f"gcloud iam service-accounts keys create \ ")
#     pprint_info(f"{str(sa_key_path)} \ ")
#     pprint_info(f"--iam-account='[SERVICE_ACCOUNT_EMAIL]'")
#
#     pprint_subheading(
#         "\nIII. Update workspace configuration with GCP Project information"
#     )
#     pprint_info(f"\nConfig file to be updated: {str(ws_config_file_path)}\n")
#
#     prereq_complete = input("Press Enter when done: ")
#     pprint_info("")
#     if not sa_key_path.exists() or not sa_key_path.is_file():
#         pprint_error(f"Could not find key at {str(sa_key_path)}, please try again")
#         return sa_key
#
#     sa_key = json.load(sa_key_path.open())
#     print(f"Key type: {type(sa_key)}")
#     print(f"Key: {sa_key}")
#
#     if sa_key is None:
#         pprint_error(
#             f"Could not read key at {str(sa_key_path)}, please verify key exists and try again"
#         )
#         return None
#
#     sa_key_type: Optional[str] = sa_key.get("type", None)
#     sa_key_project_id: Optional[str] = sa_key.get("project_id", None)
#     sa_key_client_email: Optional[str] = sa_key.get("client_email", None)
#     print(f"sa_key_type: {sa_key_type}")
#     print(f"sa_key_project_id: {sa_key_project_id}")
#
#     if sa_key_type != "service_account":
#         pprint_error(
#             f"Unexpected `type` in Service Account Key, found {sa_key_type} expected 'service_account'"
#         )
#         return None
#
#     # set the ws_gcp_data before returning the key
#     # since this could be the first time we're creating PhiGcpData for this ws
#     # we provide the project_id as well
#     config.update_ws_gcp_data(
#         ws_name=ws_name, gcp_project_id=sa_key_project_id, gcp_svc_account_key=sa_key
#     )
#     return sa_key
#
#
# def authenticate_gcp_project_for_ws(
#     ws_schema: schemas.WorkspaceSchema, config: PhiConf
# ) -> Optional[schemas.GCPProjectSchema]:
#     """Handles Authenticating a GCP project for the WorkspaceSchema.
#     Results in using the existing GCPProjectSchema for the ws OR creating a new schemas.GCPProjectSchema
#     Requires:
#     * schemas.WorkspaceSchema is valid
#     * PhiConf is valid
#     * Pak8Conf is valid
#
#     Steps:
#     1. Validate the ws_data, ws_dir_path and ws_config_file_path
#     2. Validate that the Pak8Conf has `gcp` key set.
#     3. Check if an existing GCPProjectSchema exists for this workspace,
#         if yes, check if a GKEClusterSchema is available as well. Return the existing GCPProjectSchema
#     4. Create a new GCPProjectSchema for this workspace using a service account key
#     5. Check if an existing key is available on the users machine,
#         if yes, ask the user if they'd like to use that key instead
#     6. If no key is available, or if the user would like to create a new key
#         Ask the user how they would like to create the service account and key
#         Option 1: phi creates the service account and key for the user
#         Option 2: user manually creates the key
#     7. Using the service account and key, upload the data to zeus and get a GCPProjectSchema in return.
#         Return this GCPProjectSchema
#
#     Returns
#     schemas.GCPProjectSchema: The GCPProjectSchema received from zeus
#     """
#
#     if ws_schema is None or config is None:
#         return None
#     ws_name: str = ws_schema.name
#     id_workspace: int = ws_schema.id_workspace
#     logger.debug(f"Authenticating GCPProjectSchema for {ws_name}")
#
#     ######################################################
#     # Validate WorkspaceSchema Data
#     ######################################################
#     # Step 1: Validate the ws_data, ws_dir_path and ws_config_file_path
#     ws_data: Optional[PhiWsData] = config.get_ws_data_by_name(ws_name)
#     if ws_data is None:
#         pprint_error(
#             f"WorkspaceSchema {ws_name} not registered with `phi` on this machine, please run `phi ws setup -w {ws_name}`"
#         )
#         return None
#
#     # Validate that this workspace has a ws_dir_path and ws_config_file_path set
#     # Because without the ws_dir_path and ws_config_file_path there's not point in reading the config
#     ws_dir_path: Optional[Path] = ws_data.ws_dir_path
#     ws_config_file_path: Optional[Path] = ws_data.ws_config_file_path
#     if ws_dir_path is None or ws_config_file_path is None:
#         pprint_error(
#             f"WorkspaceSchema {ws_name} has incomplete data, please run `phi ws setup -w {ws_name}` to fix"
#         )
#         return None
#
#     # Step 2: Validate that the Pak8Conf has `gcp` key set.
#     ws_pak8_conf: Optional[Pak8Conf] = config.get_ws_pak8_conf_by_name(
#         ws_name=ws_name, refresh=True
#     )
#     if ws_pak8_conf is None:
#         pprint_error("WorkspaceSchema Config invalid")
#         pprint_status("Please fix your workspace config and try again")
#         return None
#     if ws_pak8_conf.gcp is None:
#         pprint_error("This workspace does not have a GCP project configured")
#         pprint_status("Please set the `gcp` section of your ws config")
#         return None
#
#     # Step 3: Check if an existing GCPProjectSchema exists for this workspace,
#     # if yes, check if a GKEClusterSchema is available as well.
#     # Add these objects to the PhidatConfig
#     gcp_project_from_api: Optional[schemas.GCPProjectSchema] = get_primary_gcp_project(
#         id_workspace
#     )
#     if gcp_project_from_api:
#         pprint_status("Found an existing GCPProjectSchema for this workspace")
#         # since this could be the first time we're creating a PhiGcpData for this ws
#         # we set the gcp_project_id and gcp_project for this workspace
#         config.update_ws_gcp_data(
#             ws_name=ws_name,
#             gcp_project_id=gcp_project_from_api.gcp_project_id,
#             gcp_project=gcp_project_from_api,
#         )
#         gke_cluster: Optional[
#             schemas.GKEClusterSchema
#         ] = get_gke_cluster_for_gcp_project(
#             id_workspace=id_workspace,
#             id_project=gcp_project_from_api.id_project,
#         )
#         if gke_cluster is not None:
#             pprint_status("Found an existing GKEClusterSchema for this workspace")
#             config.update_ws_gcp_data(ws_name=ws_name, gke_cluster=gke_cluster)
#
#     # Step 4: Create a new GCPProjectSchema for this workspace using a service account key
#     _sa_key_path: Path = (
#         ws_pak8_conf.gcp.credentials.service_account_key_path
#         if ws_pak8_conf.gcp.credentials
#         and ws_pak8_conf.gcp.credentials.service_account_key_path
#         else get_default_sa_key_path_(ws_dir_path)
#     )
#
#     # Step 5: Check if an existing key is available on the users machine,
#     # if yes, ask the user if they'd like to use that key instead
#     _sa_key_json: Optional[Dict[str, Any]] = (
#         ws_pak8_conf.gcp.credentials.service_account_key_json
#         if ws_pak8_conf.gcp.credentials
#         and ws_pak8_conf.gcp.credentials.service_account_key_json
#         else None
#     )
#     if _sa_key_json is not None:
#         # If it does, ask the user if they'd like to use it
#         pprint_status(f"phi found a key at {_sa_key_path}")
#         use_existing_key: bool = typer.confirm(
#             "Would you like to use the existing key?"
#         )
#         if use_existing_key:
#             config.update_ws_gcp_data(
#                 ws_name=ws_name,
#                 gcp_project_id=ws_pak8_conf.gcp.project_id,
#                 gcp_svc_account_key=_sa_key_json,
#             )
#             return _upsert_primary_gcp_project_for_ws(ws_name, config)
#
#     ######################################################
#     # Create a GCP Service Account and Key for phidata
#     ######################################################
#     # Step 6: If no key is available, or if the user would like to create a new key
#     # Ask the user how they would like to create the service account and key
#     # Option 1: phi creates the service account and key for the user
#     # Option 2: user manually creates the key
#     phi_sa_name = "phidata-{}-sa".format(ws_name)
#     phi_sa_displayname = "phidata-sa"
#
#     pprint_heading("Using Phidata with Google Cloud")
#     pprint_info(
#         "To work with GCP, phi needs a Service Account to manage your resources."
#     )
#     pprint_info(
#         "Please select how you would like to create the service account and key:"
#     )
#     pprint_info(
#         "  [1] Authorize phi to use your local gcloud to create a service account and key for you (Recommended)"
#     )
#     pprint_info("  [2] Manually create the service account and key")
#     _create_sa_inp_raw: str = input("--> ")
#     _create_sa_inp: Optional[int] = str_to_int(_create_sa_inp_raw)
#     pprint_info("")
#     if _create_sa_inp is None:
#         pprint_info("\nExiting...\n")
#         return None
#
#     sa_key: Optional[Dict[str, Any]] = None
#     # NOTE: Both of these methods are expected to create the PhiGcpData for this workspace
#     # The following _upsert_gcp_project_for_ws() function only updates the gcp_project
#     if _create_sa_inp == 1:
#         sa_key = create_service_account_key_application_default_creds(
#             ws_name,
#             config,
#             phi_sa_name,
#             phi_sa_displayname,
#             _sa_key_path,
#         )
#     elif _create_sa_inp == 2:
#         sa_key = create_service_account_key_manually(
#             ws_name,
#             config,
#             phi_sa_name,
#             phi_sa_displayname,
#             _sa_key_path,
#             ws_config_file_path,
#         )
#     else:
#         pprint_error("Sorry, could not parse input. Please run `phi gcp auth` again.")
#
#     if sa_key is None:
#         pprint_error(
#             "Service Account creation failed, please run `phi gcp auth` again."
#         )
#
#     # Step 7: Using the service account and key, upload the data to zeus and get a GCPProjectSchema in return.
#     # Return this GCPProjectSchema
#     return _upsert_primary_gcp_project_for_ws(ws_name, config)



# import time
# from typing import Optional, cast
#
# import pydantic
#
# import pak8.exceptions as pak8_exceptions
# from pak8.conf import Pak8Conf
# from pak8.gcp.gcp_pak8 import GCPPak8
# from pak8.k8s.resources.kubeconfig import Kubeconfig
# from phi import enums, schemas, zeus_api
# from phi.gcp.gke.gke_utils import (
#     delete_gke_cluster_for_pak8,
#     get_or_create_gke_cluster_for_pak8,
# )
# from phi.conf.phi_conf import PhiConf, PhiGcpData, PhiK8sData, WsReleases
# from phi.utils.common import (
#     pprint_info,
#     pprint_error,
#     pprint_info,
#     pprint_status,
#     pprint_subheading,
# )
# from phi.utils.dttm import current_datetime_utc
# from phi.utils.log import logger
# from phi.zeus_api.gcp import get_gke_cluster_for_gcp_project
#
#
# def secho_gke_cluster_status(
#     gke_cluster: Optional[schemas.GKEClusterSchema] = None,
# ) -> None:
#     if gke_cluster is None:
#         pprint_error(
#             "No GKEClusterSchema available, you can create one using `phi ws up` or `phi gke create-cluster`"
#         )
#         return
#     else:
#         pprint_subheading("GKEClusterSchema:")
#         logger.debug("GKEClusterSchema: {}".format(gke_cluster.json(indent=2)))
#
#
# def print_gke_cluster_status(
#     ws_schema: schemas.WorkspaceSchema, config: PhiConf
# ) -> None:
#     """Print the GKEClusterSchema status for a workspace if available
#     This function will first validate and sync the Pak8 with zeus
#     """
#
#     # Step 1: Get the Pak8 Conf for this workspace
#     # The refresh=True will reprocess the ws config
#     pak8_conf: Optional[Pak8Conf] = config.get_ws_pak8_conf_by_name(
#         ws_schema.name, refresh=True
#     )
#     if pak8_conf is None:
#         pprint_error(f"Unable to parse config for {ws_schema.name}")
#         return
#
#     # Step 2: Sync the Pak8Conf with Zeus
#     pprint_info("Syncing WorkspaceSchema Config")
#     synced_ws: Optional[schemas.WorkspaceSchema] = zeus_api.sync_ws_pak8_config(
#         id_workspace=ws_schema.id_workspace, pak8_conf=pak8_conf
#     )
#     if synced_ws is not None:
#         pprint_info("WorkspaceSchema Sync'd")
#
#     # Step 3: Get the GCPProjectSchema for this workspace
#     gcp_project: Optional[
#         schemas.GCPProjectSchema
#     ] = config.get_gcp_project_schema_by_ws_name(ws_schema.name)
#     if gcp_project is None:
#         pprint_error(
#             "No GCP Project available, please run `phi ws setup` to setup a GCP Project for this workspace"
#         )
#         return
#
#     # Step 4: Get the GKEClusterSchema from zeus
#     pprint_info("Getting GKE Cluster status")
#     pprint_info("TO BE IMPLEMENTED: GET THE GKECLUSTER STATUS FROM GCP")
#     id_ws = gcp_project.id_workspace
#     id_prj = gcp_project.id_project
#     gke_cluster: Optional[schemas.GKEClusterSchema] = None
#     if id_ws and id_prj:
#         gke_cluster = get_gke_cluster_for_gcp_project(
#             id_workspace=id_ws,
#             id_project=id_prj,
#         )
#         if gke_cluster is not None:
#             config.update_ws_gcp_data(ws_name=ws_schema.name, gke_cluster=gke_cluster)
#             secho_gke_cluster_status(gke_cluster)
#
#
# def get_or_create_gke_cluster(
#     ws_schema: schemas.WorkspaceSchema, config: PhiConf, skip_cache: bool = False
# ) -> Optional[schemas.GKEClusterSchema]:
#     """Checks if a GKEClusterSchema for a workspace is available, if yes, returns that cluster
#     If no cluster is available, creates a GKEClusterSchema for the workspace
#     """
#
#     ws_name: str = ws_schema.name
#     logger.debug(f"Getting GKEClusterSchema for ws: {ws_name}")
#
#     # Step 1: Get the Pak8Conf for the workspace
#     # Refresh the Pak8Conf to read new changes
#     pprint_status("Reading the latest config")
#     ws_pak8_conf: Optional[Pak8Conf] = config.get_ws_pak8_conf_by_name(
#         ws_name, refresh=True
#     )
#     if ws_pak8_conf is None:
#         pprint_error("WorkspaceSchema Config invalid")
#         pprint_status("Please fix your workspace config and try again")
#         return None
#     if ws_pak8_conf.gcp is None:
#         pprint_error("This workspace does not have a GCP project configured")
#         return None
#     if ws_pak8_conf.gcp.gke_cluster_config is None:
#         pprint_error(
#             "Invalid GKE Config, please fix the gcp.gke section of your config"
#         )
#         return None
#
#     # Step 2: Get the PhiGcpData, GCPProjectSchema and check if a GKEClusterSchema is already available.
#     # If not, then we create a new GKE_CLUSTER
#     ws_gcp_data: Optional[PhiGcpData] = config.get_ws_gcp_data_by_name(ws_name)
#     if ws_gcp_data is None or ws_gcp_data.gcp_project is None:
#         pprint_error(
#             "No GCP Project available, please run `phi gcp auth` to setup a GCP Project for this workspace"
#         )
#         return None
#     # Get the GCPProjectSchema for this workspace
#     # logger.debug("Getting GCPProjectSchema for this workspace")
#     gcp_project_id: str = ws_gcp_data.gcp_project_id
#     ws_gcp_project: schemas.GCPProjectSchema = ws_gcp_data.gcp_project
#     # Check if a GKEClusterSchema is already available, if not create a new GKE_CLUSTER
#     logger.debug("Checking if a GKEClusterSchema is already available")
#     ws_gke_cluster: Optional[schemas.GKEClusterSchema] = ws_gcp_data.gke_cluster
#     # logger.debug("GKE Cluster: {}".format(ws_gke_cluster.json()))
#     if ws_gke_cluster is not None and ws_gke_cluster.is_valid and not skip_cache:
#         pprint_info(
#             f"GKEClusterSchema: {ws_gke_cluster.cluster_name} is already available and active"
#         )
#         return ws_gke_cluster
#
#     ## CREATING A NEW GKE_CLUSTER
#     pprint_info(
#         f"Creating GKEClusterSchema: {ws_pak8_conf.gcp.gke_cluster_config.name}"
#     )
#
#     # Step 3: Get the GCPPak8 for this workspace
#     ws_gcp_pak8: Optional[GCPPak8] = None
#     try:
#         ws_gcp_pak8 = GCPPak8(ws_pak8_conf)
#         # logger.debug("GCPPak8 Created")
#     except pydantic.ValidationError as e:
#         error_msg = "Config Validation Error. Details:\n{}".format(e)
#         logger.error(error_msg)
#         return None
#     # If a GCPPak8 cannot be created then return None
#     if ws_gcp_pak8 is None or not isinstance(ws_gcp_pak8, GCPPak8):
#         logger.error("Invalid GCPPak8: {}".format(ws_gcp_pak8))
#         return None
#     # We need the user details to create the cluster
#     user: schemas.UserSchema = config.user
#
#     # Step 4: Check if there is an existing CREATE_GKE_CLUSTER releases available
#     # If not, create a new CREATE_GKE_CLUSTER release
#     # TODO: Ensure there is max 1 active release
#     ws_releases: WsReleases = cast(WsReleases, config.get_ws_releases_by_name(ws_name))
#     create_gke_cluster_release: Optional[schemas.ReleaseSchema] = None
#     if ws_releases is not None:
#         existing_create_gke_cluster_releases = ws_releases.get_releases_by_type(
#             enums.ReleaseType.CREATE_GKE_CLUSTER
#         )
#         if existing_create_gke_cluster_releases:
#             logger.debug(
#                 "Found {} active {} releases".format(
#                     len(existing_create_gke_cluster_releases),
#                     enums.ReleaseType.CREATE_GKE_CLUSTER.value,
#                 )
#             )
#             for rls in existing_create_gke_cluster_releases:
#                 # logger.debug(f"id_release: {rls.id_release}")
#                 if (
#                     rls.release_data
#                     and rls.release_data.ws_gcp_project__gcp_project_id
#                     == gcp_project_id
#                 ):
#                     logger.debug(f"Using active release: {rls.id_release}")
#                     create_gke_cluster_release = rls
#                     break
#
#     # Create a CREATE_GKE_CLUSTER release if needed
#     _ws_release_data: Optional[schemas.ReleaseData] = None
#     if create_gke_cluster_release is None:
#         logger.debug("Creating new CREATE_GKE_CLUSTER ReleaseSchema")
#         _ws_release_data = schemas.ReleaseData(
#             ws__id_workspace=ws_schema.id_workspace,
#             ws_gcp_project__id_project=ws_gcp_project.id_project,
#             # This is the same as ws_gcp_project.gcp_project_id
#             ws_gcp_project__gcp_project_id=gcp_project_id,
#             ws_gke_cluster__id_cluster=None,
#             ws_gke_cluster__cluster_name=None,
#         )
#         create_gke_cluster_release = schemas.ReleaseSchema(
#             id_release="{}_{}_{}_{}".format(
#                 enums.ReleaseType.CREATE_GKE_CLUSTER.value,
#                 gcp_project_id,
#                 ws_schema.id_workspace,
#                 int(time.time()) % 1000,  # Get the last 3 digits
#             ),
#             id_workspace=ws_schema.id_workspace,
#             id_user=user.id_user,
#             is_valid=True,
#             is_at_desired_state=False,
#             release_type=enums.ReleaseType.CREATE_GKE_CLUSTER.value,
#             release_data=_ws_release_data,
#             start_ts=current_datetime_utc(),
#             end_ts=None,
#             at_desired_state_ts=None,
#             last_update_ts=None,
#         )
#         # Final check that we have a working ReleaseSchema
#         if create_gke_cluster_release is None:
#             error_msg = f"Could not create a release for project: {gcp_project_id}, please try again"
#             logger.error(error_msg)
#             return None
#         # Add release to config
#         ws_releases.add_release(create_gke_cluster_release)
#     else:
#         _ws_release_data = create_gke_cluster_release.release_data
#     # Casting _ws_release_data to schemas.ReleaseData for the typechecker
#     # Documentation: https://docs.python.org/3/library/typing.html#typing.cast
#     _ws_release_data = cast(schemas.ReleaseData, _ws_release_data)
#
#     # Now that a CREATE_GKE_CLUSTER is available, mark all existing DELETE_GKE_CLUSTER
#     # releases as complete.
#     ws_releases.set_active_releases_as_complete(enums.ReleaseType.DELETE_GKE_CLUSTER)
#
#     # Step 5: Create a GKEClusterSchema if needed
#     try:
#         ws_gke_cluster = get_or_create_gke_cluster_for_pak8(
#             user=user,
#             release_data=_ws_release_data,
#             ws_schema=ws_schema,
#             ws_gcp_project=ws_gcp_project,
#             ws_gcp_pak8=ws_gcp_pak8,
#             ws_gke_cluster=ws_gke_cluster,
#         )
#     except pak8_exceptions.GCPAuthException as e:
#         pprint_error("Received exception: {}".format(e))
#         pprint_info(
#             "Please run `phi gcp auth` and recreate the service account key again"
#         )
#     # Final check that we have a working GKEClusterSchema
#     if ws_gke_cluster is None:
#         return None
#     # Add the GKE_CLUSTER to the config and save the progress
#     config.update_ws_gcp_data(ws_name=ws_name, gke_cluster=ws_gke_cluster)
#     config.save_config()
#
#     # Step 6: Update the release data to reflect the current state
#     _ws_release_data.ws_gke_cluster__id_cluster = ws_gke_cluster.id_cluster
#     _ws_release_data.ws_gke_cluster__cluster_name = ws_gke_cluster.cluster_name
#     create_gke_cluster_release.release_data = _ws_release_data
#
#     # Step 7: Mark the release at desired state if the gke cluster is active on both phidata & gcp
#     # Mark the release as at desired state if the GKEClusterSchema is active and
#     # pak8 status implies that the GKEClusterSchema is available
#     pak8_status = ws_gcp_pak8.get_pak8_status()
#     logger.debug("Pak8 is {}".format(pak8_status))
#     if ws_gke_cluster.is_valid and pak8_status.k8s_cluster_is_available():
#         # Save the kubeconfig for this cluster
#         kconf_resource: Optional[Kubeconfig] = ws_gcp_pak8.get_kubeconfig()
#         if kconf_resource:
#             ws_k8s_data: Optional[PhiK8sData] = config.update_ws_k8s_data(
#                 ws_name=ws_name, kubeconfig_resource=kconf_resource
#             )
#             if ws_k8s_data:
#                 pprint_status(f"Kubeconfig saved to: {ws_k8s_data.kubeconfig_path}")
#         create_gke_cluster_release.set_at_desired_state()
#
#     config.save_config()
#     return ws_gke_cluster
#
#
# def delete_gke_cluster_if_exists(
#     ws_schema: schemas.WorkspaceSchema, config: PhiConf
# ) -> bool:
#     """Deletes the GKEClusterSchema for a workspace if available. Returns true if successfully
#     deleted.
#     """
#
#     ws_name: str = ws_schema.name
#     logger.debug(f"Deleting GKEClusterSchema for ws: {ws_name}")
#
#     # Step 1: Get the Pak8Conf for the workspace
#     # Refresh the Pak8Conf to read new changes
#     pprint_status("Reading the latest workspace configuration")
#     ws_pak8_conf: Optional[Pak8Conf] = config.get_ws_pak8_conf_by_name(
#         ws_name, refresh=True
#     )
#     if ws_pak8_conf is None:
#         pprint_error("WorkspaceSchema Config invalid")
#         pprint_status("Please fix your workspace config and try again")
#         return False
#     if ws_pak8_conf.gcp is None:
#         pprint_error("This workspace does not have a GCP project configured")
#         return False
#     if ws_pak8_conf.gcp.gke_cluster_config is None:
#         pprint_error(
#             "Invalid GKE Config, please fix the gcp.gke section of your config"
#         )
#         return False
#
#     # Step 2: Get the PhiGcpData, GCPProjectSchema and check if a GKEClusterSchema is available.
#     ws_gcp_data: Optional[PhiGcpData] = config.get_ws_gcp_data_by_name(ws_name)
#     if ws_gcp_data is None or ws_gcp_data.gcp_project is None:
#         pprint_error(
#             "No GCP Project available, please run `phi gcp auth` to setup a GCP Project for this workspace"
#         )
#         return False
#     # Get the GCPProjectSchema for this workspace
#     logger.debug("Getting GCPProjectSchema for this workspace")
#     gcp_project_id: str = ws_gcp_data.gcp_project_id
#     ws_gcp_project: schemas.GCPProjectSchema = ws_gcp_data.gcp_project
#     # Check if a GKEClusterSchema is already available
#     logger.debug("Checking if a GKEClusterSchema is already available")
#     ws_gke_cluster: Optional[schemas.GKEClusterSchema] = ws_gcp_data.gke_cluster
#     if ws_gke_cluster is None:
#         pprint_info("No GKEClusterSchema available to delete")
#         return True
#
#     ## DELETE THE GKE_CLUSTER
#     pprint_info(
#         f"Deleting GKEClusterSchema: {ws_pak8_conf.gcp.gke_cluster_config.name}"
#     )
#
#     # Step 3: Get the GCPPak8 for this workspace
#     ws_gcp_pak8: Optional[GCPPak8] = None
#     try:
#         ws_gcp_pak8 = GCPPak8(ws_pak8_conf)
#         logger.debug("GCPPak8 Created")
#     except pydantic.ValidationError as e:
#         error_msg = "Config Validation Error. Details:\n{}".format(e)
#         logger.error(error_msg)
#         return False
#     # If a GCP Pak8 cannot be created then return False
#     if ws_gcp_pak8 is None or not isinstance(ws_gcp_pak8, GCPPak8):
#         logger.error("Invalid GCPPak8: {}".format(ws_gcp_pak8))
#         return False
#     # We need the user details to delete the cluster
#     user: schemas.UserSchema = config.user
#
#     # Step 4: Check if there is an existing DELETE_GKE_CLUSTER releases available
#     # TODO: Ensure there is max 1 active release
#     ws_releases: WsReleases = cast(WsReleases, config.get_ws_releases_by_name(ws_name))
#     delete_gke_cluster_release: Optional[schemas.ReleaseSchema] = None
#     if ws_releases is not None:
#         existing_delete_gke_cluster_releases = ws_releases.get_releases_by_type(
#             enums.ReleaseType.DELETE_GKE_CLUSTER
#         )
#         if existing_delete_gke_cluster_releases:
#             logger.debug(
#                 "Found {} active {} releases".format(
#                     len(existing_delete_gke_cluster_releases),
#                     enums.ReleaseType.DELETE_GKE_CLUSTER.value,
#                 )
#             )
#             for rls in existing_delete_gke_cluster_releases:
#                 # logger.debug(f"id_release: {rls.id_release}")
#                 if (
#                     rls.release_data
#                     and rls.release_data.ws_gcp_project__gcp_project_id
#                     == gcp_project_id
#                 ):
#                     logger.debug(f"Using active release: {rls.id_release}")
#                     delete_gke_cluster_release = rls
#                     break
#
#     # Create a DELETE_GKE_CLUSTER release if needed
#     _ws_release_data: Optional[schemas.ReleaseData] = None
#     if delete_gke_cluster_release is None:
#         logger.debug("Creating new DELETE_GKE_CLUSTER ReleaseSchema")
#         _ws_release_data = schemas.ReleaseData(
#             ws__id_workspace=ws_schema.id_workspace,
#             ws_gcp_project__id_project=ws_gcp_project.id_project,
#             # This is the same as ws_gcp_project.gcp_project_id
#             ws_gcp_project__gcp_project_id=gcp_project_id,
#             ws_gke_cluster__id_cluster=None,
#             ws_gke_cluster__cluster_name=None,
#         )
#         delete_gke_cluster_release = schemas.ReleaseSchema(
#             id_release="{}_{}_{}_{}".format(
#                 enums.ReleaseType.DELETE_GKE_CLUSTER.value,
#                 gcp_project_id,
#                 ws_schema.id_workspace,
#                 int(time.time()) % 1000,  # Get the last 3 digits
#             ),
#             id_workspace=ws_schema.id_workspace,
#             id_user=user.id_user,
#             is_valid=True,
#             is_at_desired_state=False,
#             release_type=enums.ReleaseType.DELETE_GKE_CLUSTER.value,
#             release_data=_ws_release_data,
#             start_ts=current_datetime_utc(),
#             end_ts=None,
#             at_desired_state_ts=None,
#             last_update_ts=None,
#         )
#         # Final check that we have a working ReleaseSchema
#         if delete_gke_cluster_release is None:
#             error_msg = f"Could not create a release for project: {gcp_project_id}, please try again"
#             logger.error(error_msg)
#             return False
#         # Add release to config
#         ws_releases.add_release(delete_gke_cluster_release)
#     else:
#         _ws_release_data = delete_gke_cluster_release.release_data
#     # Casting _ws_release_data to schemas.ReleaseData for the typechecker
#     # Documentation: https://docs.python.org/3/library/typing.html#typing.cast
#     _ws_release_data = cast(schemas.ReleaseData, _ws_release_data)
#
#     # Step 5: Delete the GKEClusterSchema if needed
#     ws_gke_cluster = delete_gke_cluster_for_pak8(
#         release_data=_ws_release_data,
#         ws_gcp_pak8=ws_gcp_pak8,
#         ws_gke_cluster=ws_gke_cluster,
#     )
#     # Add the GKE_CLUSTER to the config and save the progress
#     config.update_ws_gcp_data(ws_name=ws_name, gke_cluster=ws_gke_cluster)
#     config.save_config()
#
#     # Step 6: Update the release data
#     delete_gke_cluster_release.release_data = _ws_release_data
#
#     # Step 7: Mark the release at desired state if the gke cluster is deleted on both phidata & gcp
#     # Mark the release as at desired state if the GKEClusterSchema is deleted and
#     # pak8 status implies that the GKEClusterSchema is deleted
#     pak8_status = ws_gcp_pak8.get_pak8_status()
#     logger.debug("Pak8 is {}".format(pak8_status))
#     if not ws_gke_cluster.is_valid and pak8_status.k8s_cluster_is_deleted():
#         delete_gke_cluster_release.set_at_desired_state()
#
#     # Step 8: Mark existing CREATE_GKE_CLUSTER releases as complete
#     # Getting here implies that a GKEClusterSchema is deleted so we should mark any existing
#     # CREATE_GKE_CLUSTER releases as complete
#     ws_releases.set_active_releases_as_complete(enums.ReleaseType.CREATE_GKE_CLUSTER)
#     config.save_config()
#     return True



from typing import Any, Dict, List, Optional

from rich.console import Console
from rich.style import Style

from phi.utils.log import logger

console = Console()

######################################################
## Styles
# Standard Colors: https://rich.readthedocs.io/en/stable/appendix/colors.html#appendix-colors
######################################################

heading_style = Style(
    color="green",
    bold=True,
    underline=True,
)
subheading_style = Style(color="chartreuse3", bold=True)
error_style = Style(color="red")
info_style = Style()
warn_style = Style(color="magenta")


######################################################
## Print functions
######################################################


def print_heading(msg: str) -> None:
    console.print(msg, style=heading_style)


def print_subheading(msg: str) -> None:
    console.print(msg, style=subheading_style)


def print_horizontal_line() -> None:
    console.rule()


def print_info(msg: str) -> None:
    console.print(msg, style=info_style)


def print_fix(msg: str) -> None:
    console.print("FIX  : {}".format(msg), style=info_style)


def print_error(msg: str) -> None:
    logger.error("{}".format(msg))


def print_warning(msg: str) -> None:
    logger.warning(f"{msg}")


def print_dict(_dict: dict) -> None:
    from rich.pretty import pprint

    pprint(_dict)


def print_feature_not_supported_msg() -> None:
    console.print(
        "This feature is not yet supported. Please reach out to ashpreet@phidata.com for more info",
        style=info_style,
    )


def print_conf_not_available_msg() -> None:
    logger.error("PhidataConf not available, please run `phi init`")


def print_generic_error_msg() -> None:
    console.print("Something went wrong, please try again", style=error_style)


def log_ws_not_available_msg():
    logger.warning("No active workspace")
    logger.warning("Set a workspace as active using `phi set [ws_name]`")
    logger.warning("Or run `phi ws init` to create a new workspace")


def log_network_error_msg() -> None:
    console.print(
        "NetworkError. Could not reach phidata servers.",
        style=error_style,
    )


def log_server_error_msg() -> None:
    logger.debug("ServerError: could not reach phidata servers.")


def log_auth_error_msg() -> None:
    logger.debug("AuthError: could not authenticate, please run `phi auth` again")


def print_active_workspace_not_available() -> None:
    print_warning("No active workspace")
    print_warning("Run `phi ws init` to create a new workspace")
    print_warning(
        "Run `phi ws setup` from an existing workspace directory to setup the workspace"
    )
    print_warning("Or set an existing workspace as active using `phi set [ws_name]`")


def print_available_workspaces(avl_ws_list) -> None:
    avl_ws_names = [w.name for w in avl_ws_list] if avl_ws_list else []
    print_info("Available Workspaces:\n  - {}".format("\n  - ".join(avl_ws_names)))


def get_validation_error_loc(validation_error: Dict[str, Any]) -> Optional[str]:
    if "loc" not in validation_error:
        return None
    return " -> ".join(str(e) for e in validation_error["loc"])


def print_validation_errors(validation_errors: List[Dict[str, Any]]) -> None:
    """
    Pretty prints pydantic validation errors.
    TODO: pydantic validation is known to be buggy for nested models, test this function
    """
    from rich import box
    from rich.table import Column, Table

    table = Table(
        Column(header="Field", justify="center"),
        Column(header="Error", justify="center"),
        Column(header="Context", justify="center"),
        box=box.MINIMAL,
        show_lines=True,
    )

    for err in validation_errors:
        error_loc = get_validation_error_loc(err)
        _error_ctx_raw = err.get("ctx")
        error_ctx = (
            "\n".join(f"{k}: {v}" for k, v in _error_ctx_raw.items())
            if _error_ctx_raw is not None and isinstance(_error_ctx_raw, dict)
            else ""
        )
        error_msg = err.get("msg")

        if error_loc is not None:
            table.add_row(error_loc, error_msg, error_ctx)

    if table.row_count > 0:
        console.print(table)



from typing import Optional, List

from phi.conf.phi_conf import PhiWsData
from phi.utils.cli_console import (
    print_error,
    print_info,
    print_heading,
    print_info,
)
from phi.utils.log import logger
from phi.workspace.ws_enums import WorkspaceEnv


def run_command(
    command: str,
    ws_data: PhiWsData,
    target_env: WorkspaceEnv,
    target_app: Optional[str] = None,
) -> None:
    """Run a command in devbox or databox."""

    if ws_data is None or ws_data.ws_config is None:
        print_error("WorkspaceConfig invalid")
        return
    ws_config = ws_data.ws_config

    # Final run status
    run_status: bool = False
    if target_env == WorkspaceEnv.dev:
        from phi.docker.docker_operator import run_command_docker
        from phidata.infra.docker.config import DockerConfig

        docker_configs: Optional[List[DockerConfig]] = ws_config.docker
        docker_config_to_use: Optional[DockerConfig] = None
        if docker_configs is not None and isinstance(docker_configs, list):
            if len(docker_configs) == 1:
                docker_config_to_use = docker_configs[0]
            else:
                for dc in docker_configs:
                    if dc.env == target_env:
                        docker_config_to_use = dc
                        break
        if docker_config_to_use is None:
            print_error(f"No DockerConfig found for env: {target_env.value}")
            return

        ######################################################################
        # NOTE: VERY IMPORTANT TO GET RIGHT
        # Update DockerConfig data using WorkspaceConfig
        # 1. Pass down the paths from the WorkspaceConfig
        #       These paths are used everywhere from Infra to Apps
        # 2. Pass down docker_env which is used to set the env variables
        #       when running the docker command
        ######################################################################

        # The ws_dir_path is the ROOT directory for the workspace
        docker_config_to_use.workspace_root_path = ws_data.ws_dir_path
        docker_config_to_use.workspace_config_file_path = ws_data.ws_config_file_path
        docker_config_to_use.scripts_dir = ws_config.scripts_dir
        docker_config_to_use.storage_dir = ws_config.storage_dir
        docker_config_to_use.meta_dir = ws_config.meta_dir
        docker_config_to_use.products_dir = ws_config.products_dir
        docker_config_to_use.notebooks_dir = ws_config.notebooks_dir
        docker_config_to_use.workspace_config_dir = ws_config.workspace_config_dir
        docker_config_to_use.docker_env = ws_config.docker_env

        run_status = run_command_docker(
            command=command,
            docker_config=docker_config_to_use,
            target_app=target_app,
        )

    if run_status:
        print_heading("Command run success")
    else:
        print_error("Command run failure")
